# Standard Library
import json
import logging
import urllib.parse

# Third Party
from fastapi import APIRouter, Depends, HTTPException, Query, status
from fastapi_cache.decorator import cache

# First Party
from resc_backend.constants import (
    CACHE_NAMESPACE_FINDING,
    DEFAULT_RECORDS_PER_PAGE_LIMIT,
    ERROR_MESSAGE_500,
    ERROR_MESSAGE_503,
    FINDINGS_TAG,
    REDIS_CACHE_EXPIRE,
    RWS_ROUTE_DETAILED_FINDINGS,
)
from resc_backend.db.connection import Session
from resc_backend.resc_web_service.crud import detailed_finding as detailed_finding_crud
from resc_backend.resc_web_service.dependencies import get_db_connection
from resc_backend.resc_web_service.filters import FindingsFilter
from resc_backend.resc_web_service.helpers.resc_swagger_models import Model404
from resc_backend.resc_web_service.schema import (
    detailed_finding as detailed_finding_schema,
)
from resc_backend.resc_web_service.schema.pagination_model import PaginationModel

router = APIRouter(prefix=f"{RWS_ROUTE_DETAILED_FINDINGS}", tags=[FINDINGS_TAG])
logger = logging.getLogger(__name__)


@router.get(
    "",
    response_model=PaginationModel[detailed_finding_schema.DetailedFindingRead],
    summary="Get all detailed findings",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve all the findings"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
@cache(namespace=CACHE_NAMESPACE_FINDING, expire=REDIS_CACHE_EXPIRE)
def get_all_detailed_findings(
    skip: int = Query(default=0, ge=0),
    limit: int = Query(default=DEFAULT_RECORDS_PER_PAGE_LIMIT, ge=1),
    db_connection: Session = Depends(get_db_connection),
    query_string: str = None,
) -> PaginationModel[detailed_finding_schema.DetailedFindingRead]:
    """
        Retrieve all findings objects paginated
    - **query_string**

        A query string with the following format:
            param1=value1&param2=value2&param3=value3

        Where the possible parameters are:

            - vcs_providers [enum] of type VCSProviders, possible values are: BITBUCKET, AZURE_DEVOPS.
                Will default to all if non-specified.

            - finding_statuses [enum of type FindingStatus], possible values are:NOT_ANALYZED,FALSE_POSITIVE,
              TRUE_POSITIVE. Will default to all if non-specified.

            - rule_pack_versions of type [String]

            - rule_names of type [String]

            - rule_tags of type [String] findings in the result will have at least one of the specified tags
              for the rules

            - project_name of type String

            - repository_names of type [String]

            - scan_ids of type list Integer

            - start_date_time of type datetime with the following format: 1970-01-31T00:00:00

            - end_date_time of type datetime with the following format: 1970-01-31T00:00:00

    - **db_connection**

        Session of the database connection

    - **skip**

        Integer amount of records to skip to support pagination

    - **limit**

        Integer amount of records to return, to support pagination

    - **return** [FindingRead]

        The output will contain a PaginationModel containing the list of DetailedFinding type objects,
        or an empty list if no finding was found
    """

    parsed_query_string_params = dict(urllib.parse.parse_qsl(query_string))

    if parsed_query_string_params.get("scan_ids"):
        parsed_query_string_params["scan_ids"] = json.loads(parsed_query_string_params["scan_ids"])
    if parsed_query_string_params.get("vcs_providers"):
        parsed_query_string_params["vcs_providers"] = json.loads(
            parsed_query_string_params["vcs_providers"].replace("'", '"')
        )
    if parsed_query_string_params.get("finding_statuses"):
        parsed_query_string_params["finding_statuses"] = json.loads(
            parsed_query_string_params["finding_statuses"].replace("'", '"')
        )
    if parsed_query_string_params.get("rule_names"):
        parsed_query_string_params["rule_names"] = json.loads(
            parsed_query_string_params["rule_names"].replace("'", '"')
        )
    if parsed_query_string_params.get("rule_tags"):
        parsed_query_string_params["rule_tags"] = json.loads(parsed_query_string_params["rule_tags"].replace("'", '"'))
    if parsed_query_string_params.get("rule_pack_versions"):
        parsed_query_string_params["rule_pack_versions"] = json.loads(
            parsed_query_string_params["rule_pack_versions"].replace("'", '"')
        )

    findings_filter = FindingsFilter(**parsed_query_string_params)

    findings = detailed_finding_crud.get_detailed_findings(
        db_connection, findings_filter=findings_filter, skip=skip, limit=limit
    )
    total_findings = detailed_finding_crud.get_detailed_findings_count(db_connection, findings_filter=findings_filter)

    return PaginationModel[detailed_finding_schema.DetailedFindingRead](
        data=findings, total=total_findings, limit=limit, skip=skip
    )


@router.get(
    "/{finding_id}",
    response_model=detailed_finding_schema.DetailedFindingRead,
    summary="Fetch detailed finding by ID",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve detailed finding <finding_id>"},
        404: {"model": Model404, "description": "Finding <finding_id> not found"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
@cache(namespace=CACHE_NAMESPACE_FINDING, expire=REDIS_CACHE_EXPIRE)
def read_finding(
    finding_id: int, db_connection: Session = Depends(get_db_connection)
) -> detailed_finding_schema.DetailedFindingRead:
    """
        Retrieve detailed finding by its ID

    - **db_connection**: Session of the database connection
    - **finding_id**: ID of the finding for which details need to be fetched
    - **return**: [DetailedFindingRead]
        The output will contain the details of a finding
    """
    db_finding = detailed_finding_crud.get_detailed_finding(db_connection, finding_id=finding_id)
    if db_finding is None:
        raise HTTPException(status_code=404, detail="Finding not found")
    return db_finding



# Standard Library
from functools import reduce
from typing import List


def remap_dict_keys(input_dict: dict, transformation_map: List):
    new_keys = [val[1] for val in transformation_map]
    for old_key, new_key in transformation_map:
        create_nested_dictionary(input_dict, new_key, get_value_from_nested_dictionary(input_dict, *old_key))

    output_dict = {k: v for k, v in input_dict.items() if k in [key[0] for key in new_keys]}

    return output_dict


def create_nested_dictionary(dictionary, keys, value):
    for key in keys[:-1]:
        dictionary = dictionary.setdefault(key, {})
    if not isinstance(dictionary, dict):
        dictionary = {}
        dictionary = dictionary.setdefault(keys[-1], {})
    dictionary[keys[-1]] = value


def get_value_from_nested_dictionary(dictionary, *keys):
    return reduce(lambda d, key: d.get(key) if d else None, keys, dictionary)


def delete_keys_from_nested_dict(dict_del, lst_keys):
    if not lst_keys:
        return
    if len(lst_keys) == 1:
        del dict_del[lst_keys[0]]
    else:
        for value in dict_del.values():
            if isinstance(value, dict):
                delete_keys_from_nested_dict(value, lst_keys[1:])



# Standard Library
import logging
from datetime import datetime, timedelta
from typing import List

# Third Party
from sqlalchemy import extract, func, select, update
from sqlalchemy.engine import Row
from sqlalchemy.orm import Session

# First Party
from resc_backend.constants import (
    DEFAULT_RECORDS_PER_PAGE_LIMIT,
    MAX_RECORDS_PER_PAGE_LIMIT,
)
from resc_backend.db.model import DBaudit
from resc_backend.resc_web_service.schema.finding_status import FindingStatus
from resc_backend.resc_web_service.schema.time_period import TimePeriod

logger = logging.getLogger(__name__)

YEAR = "year"
MONTH = "month"
WEEK = "week"
DAY = "day"


def create_audit(
    db_connection: Session,
    finding_id: int,
    auditor: str,
    status: FindingStatus,
    comment: str = "",
) -> DBaudit:
    """
        Audit finding, updating the status and comment
    :param db_connection:
        Session of the database connection
    :param finding_id:
        id of the finding to audit
    :param auditor:
        identifier of the person performing the audit action
    :param status:
        audit status to set, type FindingStatus
    :param comment:
        audit comment to set
    :return: DBaudit
        The output will contain the audit that was created
    """
    # UPDATE FINDING to set is_latest to FALSE.
    db_connection.execute(update(DBaudit).where(DBaudit.finding_id == finding_id).values(is_latest=False))

    # Insert new Audit.
    db_audit = DBaudit(
        finding_id=finding_id,
        auditor=auditor,
        status=status,
        comment=comment,
        timestamp=datetime.utcnow(),
        is_latest=True,
    )
    db_connection.add(db_audit)
    db_connection.commit()
    db_connection.refresh(db_audit)

    return db_audit


def get_finding_audits(
    db_connection: Session,
    finding_id: int,
    skip: int = 0,
    limit: int = DEFAULT_RECORDS_PER_PAGE_LIMIT,
) -> List[DBaudit]:
    """
        Get Audit entries for finding
    :param db_connection:
        Session of the database connection
    :param finding_id:
        id of the finding to audit
    :param skip:
        integer amount of records to skip to support pagination
    :param limit:
        integer amount of records to return, to support pagination
    :return: [DBaudit]
        The output will contain the list of audit items for the given finding
    """
    limit_val = MAX_RECORDS_PER_PAGE_LIMIT if limit > MAX_RECORDS_PER_PAGE_LIMIT else limit
    query = select(DBaudit).where(DBaudit.finding_id == finding_id)
    query = query.order_by(DBaudit.id_.desc()).offset(skip).limit(limit_val)
    finding_audits = db_connection.execute(query).scalars().all()
    return finding_audits


def get_finding_audits_count(db_connection: Session, finding_id: int) -> int:
    """
        Get count of Audit entries for finding
    :param db_connection:
        Session of the database connection
    :param finding_id:
        id of the finding to audit
    :return: total_count
        count of audit entries
    """
    query = select(func.count(DBaudit.id_)).where(DBaudit.finding_id == finding_id)
    total_count = db_connection.execute(query).scalars().one()
    return total_count


def get_audit_count_by_auditor_over_time(db_connection: Session, weeks: int = 13) -> list[Row]:
    """
        Retrieve count audits by auditor over time for given weeks
    :param db_connection:
        Session of the database connection
    :param weeks:
        optional, filter on last n weeks, default 13
    :return: count_over_time
        list of rows containing audit count over time per week
    """
    last_nth_week_date_time = datetime.utcnow() - timedelta(weeks=weeks)

    query = (
        db_connection.query(
            extract(YEAR, DBaudit.timestamp).label(YEAR),
            extract(WEEK, DBaudit.timestamp).label(WEEK),
            DBaudit.auditor,
            func.count(DBaudit.id_).label("audit_count"),
        )
        .filter(
            (extract(YEAR, DBaudit.timestamp) > extract(YEAR, last_nth_week_date_time))
            | (
                (extract(YEAR, DBaudit.timestamp) == extract(YEAR, last_nth_week_date_time))
                & (extract(WEEK, DBaudit.timestamp) >= extract(WEEK, last_nth_week_date_time))
            )
        )
        .group_by(
            extract(YEAR, DBaudit.timestamp).label(YEAR),
            extract(WEEK, DBaudit.timestamp).label(WEEK),
            DBaudit.auditor,
        )
        .order_by(
            extract(YEAR, DBaudit.timestamp).label(YEAR),
            extract(WEEK, DBaudit.timestamp).label(WEEK),
            DBaudit.auditor,
        )
    )
    finding_audits = query.all()

    return finding_audits


def get_personal_audit_count(db_connection: Session, auditor: str, time_period: TimePeriod) -> int:
    """
        Get count of Audit entries for finding
    :param db_connection:
        Session of the database connection
    :param auditor:
        id of the auditor
    :param time_period:
        period for which to retrieve the audit counts
    :return: total_count
        count of audit entries
    """
    date_today = datetime.utcnow()

    total_count = db_connection.query(func.count(DBaudit.id_))

    if time_period in (time_period.DAY, time_period.MONTH, time_period.YEAR):
        total_count = total_count.filter(extract(YEAR, DBaudit.timestamp) == extract(YEAR, date_today))

        if time_period in (time_period.DAY, time_period.MONTH):
            total_count = total_count.filter(extract(MONTH, DBaudit.timestamp) == extract(MONTH, date_today))

            if time_period == time_period.DAY:
                total_count = total_count.filter(extract(DAY, DBaudit.timestamp) == extract(DAY, date_today))

    if time_period in (time_period.WEEK, time_period.LAST_WEEK):
        date_last_week = datetime.utcnow() - timedelta(weeks=1)
        date_week = date_last_week if time_period == time_period.LAST_WEEK else date_today
        total_count = total_count.filter(extract(YEAR, DBaudit.timestamp) == extract(YEAR, date_week))
        total_count = total_count.filter(extract(WEEK, DBaudit.timestamp) == extract(WEEK, date_week))

    total_count = total_count.filter(DBaudit.auditor == auditor).scalar()
    return total_count



# Standard Library
import logging

# Third Party
import requests
from requests.adapters import HTTPAdapter
from requests.auth import HTTPBasicAuth
from urllib3 import Retry

# First Party
from resc_backend.common import initialise_logs
from resc_backend.constants import (
    LOG_FILE_PATH_RABBITMQ,
    PROJECT_QUEUE,
    REPOSITORY_QUEUE,
)
from resc_backend.helpers.environment_wrapper import validate_environment
from resc_backend.helpers.rabbitmq.configuration import (
    RABBITMQ_DEFAULT_PASSWORD,
    RABBITMQ_DEFAULT_USER,
    RABBITMQ_DEFAULT_VHOST,
    RABBITMQ_PASSWORD,
    RABBITMQ_USERNAME,
    REQUIRED_ENV_VARS,
)

env_variables = validate_environment(REQUIRED_ENV_VARS)
logger_config = initialise_logs(LOG_FILE_PATH_RABBITMQ)
logger = logging.getLogger(__name__)


def wait_for_rabbitmq_server_to_up(rabbitmq_api_base_url: str) -> bool:
    """
        This function returns true once the RabbitMQ server is up.
    :param rabbitmq_api_base_url:
        RabbitMQ API base url
    """
    rabbitmq_admin_user = f"{env_variables[RABBITMQ_DEFAULT_USER]}"
    rabbitmq_admin_password = f"{env_variables[RABBITMQ_DEFAULT_PASSWORD]}"

    session = requests.Session()
    retries = Retry(total=100, backoff_factor=1, status_forcelist=[500, 502, 503, 504])
    uri = rabbitmq_api_base_url + "/api/health/checks/alarms"
    session.mount("http://", HTTPAdapter(max_retries=retries))
    response = session.get(uri, auth=HTTPBasicAuth(rabbitmq_admin_user, rabbitmq_admin_password))

    if hasattr(response, "status_code") and int(response.status_code) == 200:
        logger.debug("Rabbitmq server is up.")
        return True
    logger.error(f"Rabbitmq server is down, HTTP status: '{response.status_code}'")
    return False


def create_user(rabbitmq_api_base_url: str, username: str, password: str, role: str) -> bool:
    """
        Creates a user in RabbitMQ server, it would return true if the operation was a success, false otherwise
    :param rabbitmq_api_base_url:
        RabbitMQ API base url
    :param username:
        Username to be created
    :param password:
        Password of the user to be created
    :param role:
        Role of the user to be created
    :return: True if user creation is successful else returns False.
    """
    rabbitmq_admin_user = f"{env_variables[RABBITMQ_DEFAULT_USER]}"
    rabbitmq_admin_password = f"{env_variables[RABBITMQ_DEFAULT_PASSWORD]}"

    uri = rabbitmq_api_base_url + "/api/users/" + username
    user_data = {"password": password, "tags": role}
    response = requests.put(
        uri,
        json=user_data,
        auth=HTTPBasicAuth(rabbitmq_admin_user, rabbitmq_admin_password),
        timeout=10,
    )

    if hasattr(response, "status_code") and int(response.status_code) == 201:
        logger.info(f"User: {username} with role: {role} created successfully.")
        return True
    if hasattr(response, "status_code") and int(response.status_code) == 204:
        logger.info(f"User: {username} with role: {role} already exists.")
        return True
    logger.error(
        f"Failed while creating user: '{username}' " f"with role: '{role}' " f", HTTP status: '{response.status_code}'"
    )
    return False


def set_resource_permissions(
    rabbitmq_api_base_url: str,
    v_host: str,
    username: str,
    configure_resources_regex: str,
    read_resources_regex: str,
    write_resources_regex: str,
) -> bool:
    """
        Sets permission for resources, it would return true if the operation was a success, false otherwise
    :param rabbitmq_api_base_url:
        RabbitMQ API base url
    :param v_host:
        Virtual host name of the RabbitMQ server
    :param username:
        User whose permission needs to set
    :param configure_resources_regex:
        Regex of the resources for which permission needs to set
    :param read_resources_regex:
        Regex of read permission
    :param write_resources_regex:
        Regex of write permission
    :return: True if permission assigned to user is successful else returns False.
    """
    rabbitmq_admin_user = f"{env_variables[RABBITMQ_DEFAULT_USER]}"
    rabbitmq_admin_password = f"{env_variables[RABBITMQ_DEFAULT_PASSWORD]}"

    uri = rabbitmq_api_base_url + "/api/permissions/" + v_host + "/" + username
    permission_data = {
        "configure": configure_resources_regex,
        "write": write_resources_regex,
        "read": read_resources_regex,
    }

    response = requests.put(
        uri,
        json=permission_data,
        auth=HTTPBasicAuth(rabbitmq_admin_user, rabbitmq_admin_password),
        timeout=10,
    )
    if hasattr(response, "status_code") and int(response.status_code) == 201:
        logger.debug(f"vHost permission successfully assigned to user: {username} for vHost: {v_host}.")
        return True
    logger.error(
        f"Failed while assigning vhost permission to user: '{username}' "
        f"for vhost: '{v_host}' "
        f", HTTP status: '{response.status_code}'"
    )
    return False


def set_topic_permissions(
    rabbitmq_api_base_url: str,
    v_host: str,
    username: str,
    topic_name: str,
    allow_read: bool,
    allow_write: bool,
) -> bool:
    """
        Sets permission for topics, it would return true if the operation was a success, false otherwise
    :param rabbitmq_api_base_url:
        RabbitMQ API base url
    :param v_host:
        Virtual host name of the RabbitMQ server
    :param username:
        User whose permission needs to set
    :param topic_name:
        Name of the topic for which permission needs to set
    :param allow_read:
        Read permission allowed or not
    :param allow_write:
        Write permission allowed or not
    :return: True if topic permission assigned to user is successful else returns False.
    """
    permission_allow = ".*"
    permission_deny = "^$"
    read_permission = permission_deny
    write_permission = permission_deny

    if allow_read:
        read_permission = permission_allow
    if allow_write:
        write_permission = permission_allow

    rabbitmq_admin_user = f"{env_variables[RABBITMQ_DEFAULT_USER]}"
    rabbitmq_admin_password = f"{env_variables[RABBITMQ_DEFAULT_PASSWORD]}"

    uri = rabbitmq_api_base_url + "/api/topic-permissions/" + v_host + "/" + username
    topic_permission_data = {
        "exchange": topic_name,
        "write": write_permission,
        "read": read_permission,
    }

    response = requests.put(
        uri,
        json=topic_permission_data,
        auth=HTTPBasicAuth(rabbitmq_admin_user, rabbitmq_admin_password),
        timeout=10,
    )

    if hasattr(response, "status_code") and (int(response.status_code) == 201 or int(response.status_code) == 204):
        logger.debug(
            f"Topic permission successfully assigned to user: {username} for topic: {topic_name} in vhost: {v_host}."
        )
        return True
    logger.error(
        f"Failed while assigning topic permission to user: '{username}' "
        f"with vhost: '{v_host}' "
        f"for topic: '{topic_name}' "
        f", HTTP status: '{response.status_code}'"
    )
    return False


def create_queue_user_and_set_permission(rabbitmq_api_base_url: str):
    """
        This function creates user for queue and assigns required permission.
    :param rabbitmq_api_base_url:
        RabbitMQ API base url
    """
    rabbitmq_vhost = f"{env_variables[RABBITMQ_DEFAULT_VHOST]}"
    queue_user = f"{env_variables[RABBITMQ_USERNAME]}"
    queue_password = f"{env_variables[RABBITMQ_PASSWORD]}"

    user_created = create_user(
        rabbitmq_api_base_url=rabbitmq_api_base_url,
        username=queue_user,
        password=queue_password,
        role="monitoring",
    )
    if user_created:
        set_resource_permissions(
            rabbitmq_api_base_url=rabbitmq_api_base_url,
            v_host=rabbitmq_vhost,
            username=queue_user,
            configure_resources_regex=f"^({PROJECT_QUEUE}|{REPOSITORY_QUEUE}" f"|.*celery.*)$",
            read_resources_regex=f"^{PROJECT_QUEUE}|{REPOSITORY_QUEUE}" f"|.*celery.*$",
            write_resources_regex=f"^{PROJECT_QUEUE}|{REPOSITORY_QUEUE}" f"|amq.default|.*celery.*$",
        )
        set_topic_permissions(
            rabbitmq_api_base_url=rabbitmq_api_base_url,
            v_host=rabbitmq_vhost,
            username=queue_user,
            topic_name=f"{PROJECT_QUEUE}",
            allow_read=True,
            allow_write=True,
        )
        set_topic_permissions(
            rabbitmq_api_base_url=rabbitmq_api_base_url,
            v_host=rabbitmq_vhost,
            username=queue_user,
            topic_name=f"{REPOSITORY_QUEUE}",
            allow_read=True,
            allow_write=True,
        )



# Standard Library
from datetime import datetime
from typing import List

# Third Party
from sqlalchemy import func, update
from sqlalchemy.orm import Session

# First Party
from resc_backend.constants import (
    DEFAULT_RECORDS_PER_PAGE_LIMIT,
    MAX_RECORDS_PER_PAGE_LIMIT,
)
from resc_backend.db.model import (
    DBfinding,
    DBrepository,
    DBscan,
    DBscanFinding,
    DBVcsInstance,
)
from resc_backend.resc_web_service.crud import finding as finding_crud
from resc_backend.resc_web_service.crud import scan_finding as scan_finding_crud
from resc_backend.resc_web_service.schema import scan as scan_schema
from resc_backend.resc_web_service.schema.finding_status import FindingStatus
from resc_backend.resc_web_service.schema.scan_type import ScanType


def get_scan(db_connection: Session, scan_id: int) -> DBscan:
    scan = db_connection.query(DBscan).where(DBscan.id_ == scan_id).first()
    return scan


def get_latest_scan_for_repository(db_connection: Session, repository_id: int) -> DBscan:
    """
        Retrieve the most recent scan of a given repository object
    :param db_connection:
        Session of the database connection
    :param repository_id:
        id of the repository object for which to retrieve the most recent scan
    :return: scan
        scan object having the most recent timestamp for a given repository object
    """
    subquery = db_connection.query(func.max(DBscan.timestamp).label("max_time"))
    subquery = subquery.where(DBscan.repository_id == repository_id)
    subquery = subquery.subquery()

    query = db_connection.query(DBscan)
    query = query.join(subquery, DBscan.timestamp == subquery.c.max_time)
    query = query.where(DBscan.repository_id == repository_id)
    scan = query.first()

    return scan


def get_scans(
    db_connection: Session,
    skip: int = 0,
    limit: int = DEFAULT_RECORDS_PER_PAGE_LIMIT,
    repository_id: int = -1,
) -> List[DBscan]:
    """
        Retrieve the scan records, ordered by scan_id and optionally filtered by repository_id
    :param db_connection:
        Session of the database connection
    :param repository_id:
        optional int filtering the repository for which to retrieve scans
    :param skip:
        integer amount of records to skip to support pagination
    :param limit:
        integer amount of records to return, to support pagination
    :return: [DBscan]
        List of DBScan objects
    """
    limit_val = MAX_RECORDS_PER_PAGE_LIMIT if limit > MAX_RECORDS_PER_PAGE_LIMIT else limit
    query = db_connection.query(DBscan)

    if repository_id > 0:
        query = query.where(DBscan.repository_id == repository_id)

    scans = query.order_by(DBscan.id_).offset(skip).limit(limit_val).all()
    return scans


def get_scans_count(db_connection: Session, repository_id: int = -1) -> int:
    """
        Retrieve count of scan records optionally filtered by VCS provider
    :param db_connection:
        Session of the database connection
    :param repository_id:
        optional int filtering the repository for which to retrieve scans
    :return: total_count
        count of scans
    """
    query = db_connection.query(func.count(DBscan.id_))

    if repository_id > 0:
        query = query.where(DBscan.repository_id == repository_id)

    total_count = query.scalar()
    return total_count


def update_scan(db_connection: Session, scan_id: int, scan: scan_schema.ScanCreate) -> DBscan:
    db_scan = db_connection.query(DBscan).filter_by(id_=scan_id).first()
    db_scan.scan_type = scan.scan_type
    db_scan.last_scanned_commit = scan.last_scanned_commit
    db_scan.timestamp = scan.timestamp
    db_scan.increment_number = scan.increment_number
    db_scan.rule_pack = scan.rule_pack
    db_connection.commit()
    db_connection.refresh(db_scan)
    return db_scan


def create_scan(db_connection: Session, scan: scan_schema.ScanCreate) -> DBscan:
    # We only flag the previous ones if we are doing a base scan.
    # In the other cases they are simply marked as latest for incremental.
    if scan.scan_type == ScanType.BASE:
        query = update(DBscan)
        query = query.where(DBscan.repository_id == scan.repository_id)
        query = query.where(DBscan.rule_pack == scan.rule_pack)
        query = query.values(is_latest=False)
        db_connection.execute(query)

    db_scan = DBscan(
        scan_type=scan.scan_type,
        last_scanned_commit=scan.last_scanned_commit,
        repository_id=scan.repository_id,
        timestamp=scan.timestamp,
        increment_number=scan.increment_number,
        rule_pack=scan.rule_pack,
        is_latest=True,
    )
    db_connection.add(db_scan)
    db_connection.commit()
    db_connection.refresh(db_scan)
    return db_scan


def get_repository_findings_metadata_for_latest_scan(
    db_connection: Session, repository_id: int, scan_timestamp: datetime
):
    """
        Retrieves the finding metadata for latest scan of a repository from the database
    :param db_connection:
        Session of the database connection
    :param repository_id:
        repository id of the latest scan
    :param scan_timestamp:
        timestamp of the latest scan
    :return: findings_metadata
        findings_metadata containing the count for each status
    """
    scan_ids_latest_to_base = []
    scans = get_scans(db_connection=db_connection, repository_id=repository_id, limit=1000000)
    scans.sort(key=lambda x: x.timestamp, reverse=True)
    for scan in scans:
        if scan.timestamp <= scan_timestamp:
            scan_ids_latest_to_base.append(scan.id_)
            if scan.scan_type == ScanType.BASE:
                break

    true_positive_count = false_positive_count = not_analyzed_count = under_review_count = (
        clarification_required_count
    ) = 0
    if len(scan_ids_latest_to_base) > 0:
        findings_count_by_status = finding_crud.get_findings_count_by_status(
            db_connection,
            scan_ids=scan_ids_latest_to_base,
            finding_statuses=FindingStatus,
        )
        for finding in findings_count_by_status:
            finding_status = finding[1]
            count = finding[0]
            if finding_status == FindingStatus.TRUE_POSITIVE.value:
                true_positive_count = count
            if finding_status == FindingStatus.FALSE_POSITIVE.value:
                false_positive_count = count
            if finding_status == FindingStatus.NOT_ANALYZED.value or finding_status is None:
                not_analyzed_count += count
            if finding_status == FindingStatus.UNDER_REVIEW.value:
                under_review_count = count
            if finding_status == FindingStatus.CLARIFICATION_REQUIRED.value:
                clarification_required_count = count

    total_findings_count = (
        true_positive_count
        + false_positive_count
        + not_analyzed_count
        + under_review_count
        + clarification_required_count
    )

    findings_metadata = {
        "true_positive": true_positive_count,
        "false_positive": false_positive_count,
        "not_analyzed": not_analyzed_count,
        "under_review": under_review_count,
        "clarification_required": clarification_required_count,
        "total_findings_count": total_findings_count,
    }

    return findings_metadata


def delete_repository_findings_not_linked_to_any_scan(db_connection: Session, repository_id: int):
    """
        Delete findings for a given repository which are not linked to any scans
    :param db_connection:
        Session of the database connection
    :param repository_id:
        id of the repository
    """
    sub_query = db_connection.query(DBscanFinding.finding_id).distinct()
    query = db_connection.query(DBfinding)
    query = query.where(DBfinding.id_.not_in(sub_query))
    query = query.where(DBfinding.repository_id == repository_id)
    query.delete(synchronize_session=False)
    db_connection.commit()


def delete_scan(
    db_connection: Session,
    repository_id: int,
    scan_id: int,
    delete_related: bool = False,
):
    """
        Delete a scan object
    :param db_connection:
        Session of the database connection
    :param repository_id:
        repository_id for which findings will be deleted which are not linked to any scans
    :param scan_id:
        id of the scan to be deleted
    :param delete_related:
        if related records need to be deleted
    """
    if delete_related:
        scan_finding_crud.delete_scan_finding(db_connection, scan_id=scan_id)

    query = db_connection.query(DBscan)
    query = query.where(DBscan.id_ == scan_id)
    query.delete(synchronize_session=False)
    db_connection.commit()

    delete_repository_findings_not_linked_to_any_scan(db_connection, repository_id=repository_id)


def delete_scans_by_repository_id(db_connection: Session, repository_id: int):
    """
        Delete scans for a given repository
    :param db_connection:
        Session of the database connection
    :param repository_id:
        id of the repository
    """
    db_connection.query(DBscan).where(DBscan.repository_id == repository_id).delete(synchronize_session=False)
    db_connection.commit()


def delete_scans_by_vcs_instance_id(db_connection: Session, vcs_instance_id: int):
    """
        Delete scans for a given vcs instance
    :param db_connection:
        Session of the database connection
    :param vcs_instance_id:
        id of the vcs instance
    """
    query = db_connection.query(DBscan)
    query = query.where(DBscan.repository_id == DBrepository.id_)
    query = query.where(DBrepository.vcs_instance == DBVcsInstance.id_)
    query = query.where(DBVcsInstance.id_ == vcs_instance_id)
    query.delete(synchronize_session=False)
    db_connection.commit()



# ruff: noqa: E402
# ruff: noqa: F401
# Standard Library
import logging
import os

# Third Party
from sqlalchemy.orm import declarative_base

Base = declarative_base()
basedir = os.path.abspath(os.path.dirname(__file__))
logger = logging.getLogger(__name__)

# First Party
from resc_backend.db.model.audit import DBaudit
from resc_backend.db.model.finding import DBfinding
from resc_backend.db.model.repository import DBrepository
from resc_backend.db.model.rule import DBrule
from resc_backend.db.model.rule_allow_list import DBruleAllowList
from resc_backend.db.model.rule_pack import DBrulePack
from resc_backend.db.model.rule_tag import DBruleTag
from resc_backend.db.model.scan import DBscan
from resc_backend.db.model.scan_finding import DBscanFinding
from resc_backend.db.model.tag import DBtag
from resc_backend.db.model.vcs_instance import DBVcsInstance



# Standard Library
import logging
from typing import Optional

# Third Party
from fastapi import APIRouter, Depends, HTTPException, Query, status
from sqlalchemy.exc import IntegrityError

# First Party
from resc_backend.constants import (
    CACHE_NAMESPACE_VCS_INSTANCE,
    DEFAULT_RECORDS_PER_PAGE_LIMIT,
    ERROR_MESSAGE_500,
    ERROR_MESSAGE_503,
    RWS_ROUTE_VCS,
    VCS_TAG,
)
from resc_backend.db.connection import Session
from resc_backend.resc_web_service.cache_manager import CacheManager
from resc_backend.resc_web_service.crud import vcs_instance as vcs_instance_crud
from resc_backend.resc_web_service.dependencies import get_db_connection
from resc_backend.resc_web_service.helpers.resc_swagger_models import Model400, Model404
from resc_backend.resc_web_service.schema import vcs_instance as vcs_instance_schema
from resc_backend.resc_web_service.schema.pagination_model import PaginationModel
from resc_backend.resc_web_service.schema.vcs_provider import VCSProviders

router = APIRouter(prefix=f"{RWS_ROUTE_VCS}", tags=[VCS_TAG])
logger = logging.getLogger(__name__)


@router.post(
    "",
    response_model=vcs_instance_schema.VCSInstanceRead,
    summary="Create a VCS instance",
    status_code=status.HTTP_201_CREATED,
    responses={
        201: {"description": "Create new VCS instance"},
        400: {"model": Model400, "description": "Error creating new VCS instance"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
async def create_vcs_instance(
    vcs_instance: vcs_instance_schema.VCSInstanceCreate,
    db_connection: Session = Depends(get_db_connection),
) -> vcs_instance_schema.VCSInstanceRead:
    """
        Create new VCS instance object

    - **db_connection**: Session of the database connection
    - **vcs_instance**: VCSInstanceCreate type object of the VCS Instance to create
    - **return**: VCSInstanceRead
        The output will contain a VCSInstanceRead type object if the creation was successful
    """
    try:
        created_vcs_instance = vcs_instance_crud.create_vcs_instance_if_not_exists(
            db_connection=db_connection, vcs_instance=vcs_instance
        )
    except IntegrityError as err:
        logger.debug(f"Error creating new vcs_instance object: {err}")
        raise HTTPException(status_code=400, detail="Error creating new vcs_instance object") from err
    vcs_instance = vcs_instance_schema.VCSInstanceRead.create_from_db_vcs_instance(created_vcs_instance)

    # Clear cache related to VCS instances
    await CacheManager.clear_cache_by_namespace(namespace=CACHE_NAMESPACE_VCS_INSTANCE)
    return vcs_instance


@router.get(
    "/{vcs_instance_id}",
    response_model=vcs_instance_schema.VCSInstanceRead,
    summary="Fetch a VCS instance by ID",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve VCS Instance <vcs_instance_id>"},
        404: {
            "model": Model404,
            "description": "VCS Instance <vcs_instance_id> not found",
        },
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
def read_vcs_instance(
    vcs_instance_id: int, db_connection: Session = Depends(get_db_connection)
) -> vcs_instance_schema.VCSInstanceRead:
    """
        Retrieve a VCS instance object based on the provided id

    - **db_connection**: Session of the database connection
    - **vcs_instance_id**: ID of the VCS instance for which details need to be fetched
    - **return**: VCSInstanceRead
        The output will contain a VCSInstanceRead type object from the requested ID
    """
    db_vcs_instance = vcs_instance_crud.get_vcs_instance(db_connection, vcs_instance_id=vcs_instance_id)
    if db_vcs_instance is None:
        raise HTTPException(status_code=404, detail="VCS Instance not found")
    vcs_instance = vcs_instance_schema.VCSInstanceRead.create_from_db_vcs_instance(db_vcs_instance)
    return vcs_instance


@router.get(
    "",
    response_model=PaginationModel[vcs_instance_schema.VCSInstanceRead],
    summary="Get all VCS instances",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve all the VCS Instances"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
def get_all_vcs_instances(
    skip: int = Query(default=0, ge=0),
    limit: int = Query(default=DEFAULT_RECORDS_PER_PAGE_LIMIT, ge=1),
    vcs_provider_type: Optional[VCSProviders] = Query(None),
    vcs_instance_name: Optional[str] = Query(None),
    db_connection: Session = Depends(get_db_connection),
) -> PaginationModel[vcs_instance_schema.VCSInstanceRead]:
    """
        Retrieve all VCS instance objects paginated

    - **db_connection**: Session of the database connection
    - **skip**: Integer amount of records to skip to support pagination
    - **limit**: Integer amount of records to return, to support pagination
    - **vcs_provider_type**: Optional filter on the VCS provider type
    - **vcs_instance_name**: Optional filter on VCS instance name
    - **return**: [VCSInstanceRead]
        The output will contain a PaginationModel containing the list of VCSInstanceRead type objects,
        or an empty list if no VCS instance was found
    """
    db_vcs_instances = vcs_instance_crud.get_vcs_instances(
        db_connection,
        skip=skip,
        limit=limit,
        vcs_provider_type=vcs_provider_type,
        vcs_instance_name=vcs_instance_name,
    )
    vcs_instances = []
    for db_vcs_instance in db_vcs_instances:
        vcs_instances.append(vcs_instance_schema.VCSInstanceRead.create_from_db_vcs_instance(db_vcs_instance))

    total_vcs_instances = vcs_instance_crud.get_vcs_instances_count(
        db_connection,
        vcs_provider_type=vcs_provider_type,
        vcs_instance_name=vcs_instance_name,
    )

    return PaginationModel[vcs_instance_schema.VCSInstanceRead](
        data=vcs_instances, total=total_vcs_instances, limit=limit, skip=skip
    )


@router.put(
    "/{vcs_instance_id}",
    response_model=vcs_instance_schema.VCSInstanceRead,
    summary="Update an existing VCS instance",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Update VCS Instance <vcs_instance_id>"},
        404: {
            "model": Model404,
            "description": "VCS Instance <vcs_instance_id> not found",
        },
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
async def update_vcs_instance(
    vcs_instance_id: int,
    vcs_instance: vcs_instance_schema.VCSInstanceCreate,
    db_connection: Session = Depends(get_db_connection),
) -> vcs_instance_schema.VCSInstanceRead:
    """
        Update a VCS instance

    - **db_connection**: Session of the database connection
    - **vcs_instance_id**: ID of the VCS instance to update
    - **provider_type**: VCS instance name that needs to be updated
    - **hostname**: Host name of the VCS instance that needs to be updated
    - **port**: Port number of the VCS instance that needs to be updated
    - **scheme**: Scheme of the VCS instance that needs to be updated. Allowed values http or https
    - **exceptions**: List of projects which needs to be updated to exception list, default empty list
    - **scope**: List of projects which needs to be updated to scope
    - **organization**: Name of organization to be updated, default is empty
    - **return**: VCSInstanceRead
        The output will contain a VCSInstanceRead type object with the new properties
    """
    db_vcs_instance = vcs_instance_crud.get_vcs_instance(db_connection, vcs_instance_id=vcs_instance_id)
    if db_vcs_instance is None:
        raise HTTPException(status_code=404, detail="VCS instance not found")
    db_vcs_instance = vcs_instance_crud.update_vcs_instance(
        db_connection=db_connection,
        vcs_instance_id=vcs_instance_id,
        vcs_instance=vcs_instance,
    )
    vcs_instance = vcs_instance_schema.VCSInstanceRead.create_from_db_vcs_instance(db_vcs_instance)

    # Clear cache related to VCS instances
    await CacheManager.clear_cache_by_namespace(namespace=CACHE_NAMESPACE_VCS_INSTANCE)
    return vcs_instance


@router.delete(
    "/{vcs_instance_id}",
    summary="Delete a VCS instance",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Delete VCS Instance <vcs_instance_id>"},
        404: {
            "model": Model404,
            "description": "VCS Instance <vcs_instance_id> not found",
        },
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
async def delete_vcs_instance(vcs_instance_id: int, db_connection: Session = Depends(get_db_connection)):
    """
        Delete a VCS instance by ID

    - **db_connection**: Session of the database connection
    - **vcs_instance_id**: ID of the VCS instance to delete
    - **return**: The output will contain a success or error message based on the success of the deletion
    """
    db_vcs_instance = vcs_instance_crud.get_vcs_instance(db_connection, vcs_instance_id=vcs_instance_id)
    if db_vcs_instance is None:
        raise HTTPException(status_code=404, detail="VCS instance not found")
    vcs_instance_crud.delete_vcs_instance(db_connection, vcs_instance_id=vcs_instance_id, delete_related=True)

    # Clear cache related to VCS instances
    await CacheManager.clear_cache_by_namespace(namespace=CACHE_NAMESPACE_VCS_INSTANCE)
    return {"ok": True}



# Standard Library
from typing import Optional

# Third Party
from pydantic import BaseModel, conint, constr


class RuleBase(BaseModel):
    rule_name: constr(min_length=1, max_length=400)
    description: Optional[constr(max_length=2000)] = None
    entropy: Optional[float] = None
    secret_group: Optional[int] = None
    regex: Optional[str] = None
    path: Optional[str] = None
    keywords: Optional[str] = None


class RuleCreate(RuleBase):
    rule_pack: constr(regex=r"^(\d+\.)?(\d+\.)?(\*|\d+)$")
    allow_list: Optional[conint(gt=0)] = None

    @classmethod
    def create_from_base_class(cls, base_object: RuleBase, rule_pack: str, allow_list=int):
        return cls(**(dict(base_object)), rule_pack=rule_pack, allow_list=allow_list)


class Rule(RuleBase):
    pass


class RuleRead(RuleCreate):
    id_: conint(gt=0)

    class Config:
        orm_mode = True



# Standard Library
import logging
import ssl
import urllib.error

# Third Party
import jwt
import sqlalchemy.orm
from fastapi import Depends, HTTPException, Request, status
from fastapi.security import HTTPBasicCredentials, HTTPBearer
from jwt import PyJWKClient
from tenacity import retry, stop_after_attempt, wait_exponential

# First Party
from resc_backend.constants import (
    CACHE_CONTROL,
    CONTENT_SECURITY_POLICY,
    CROSS_ORIGIN_RESOURCE_POLICY,
    REFERRER_POLICY,
    STRICT_TRANSPORT_SECURITY,
    X_CONTENT_TYPE_OPTIONS,
    X_FRAME_OPTIONS,
    X_PERMITTED_CROSS_DOMAIN_POLICIES,
    X_XSS_PROTECTION,
)
from resc_backend.db.connection import Session, engine
from resc_backend.db.model import DBfinding, DBrepository, DBrule, DBscan, DBscanFinding
from resc_backend.helpers.environment_wrapper import validate_environment
from resc_backend.resc_web_service.configuration import (
    CONDITIONAL_SSO_ENV_VARS,
    SSO_ACCESS_TOKEN_ISSUER_URL,
    SSO_ACCESS_TOKEN_JWKS_URL,
    SSO_JWT_CLAIM_KEY_AUTHORIZATION,
    SSO_JWT_CLAIM_KEY_USER_ID,
    SSO_JWT_CLAIM_VALUE_AUTHORIZATION,
    SSO_JWT_REQUIRED_CLAIMS,
    SSO_JWT_SIGN_ALGORITHM,
)

security = HTTPBearer()
logger = logging.getLogger(__name__)


async def requires_auth(request: Request, credentials: HTTPBasicCredentials = Depends(security)):
    """
    Function that is used to validate the JWT access token
    """
    # Check and load environment variables
    env_variables = validate_environment(CONDITIONAL_SSO_ENV_VARS)

    access_token = credentials.credentials
    algorithm = [env_variables[SSO_JWT_SIGN_ALGORITHM]]
    issuer = env_variables[SSO_ACCESS_TOKEN_ISSUER_URL]
    jwks_url = env_variables[SSO_ACCESS_TOKEN_JWKS_URL]
    jwt_options = {
        "verify_signature": True,
        "verify_iss": True,
        "verify_exp": True,
        "require": env_variables[SSO_JWT_REQUIRED_CLAIMS].split(","),
    }
    try:
        ssl._create_default_https_context = ssl._create_unverified_context
        jwks_client = PyJWKClient(jwks_url)
        signing_key = jwks_client.get_signing_key_from_jwt(access_token)
        claims = jwt.decode(
            access_token,
            signing_key.key,
            algorithms=algorithm,
            issuer=issuer,
            options=jwt_options,
        )
        user_id = claims[env_variables[SSO_JWT_CLAIM_KEY_USER_ID]]
        if not user_is_authorized(claims, env_variables):
            logger.error(f"Invalid login attempt for user {user_id}")
            raise HTTPException(
                status_code=status.HTTP_403_FORBIDDEN,
                detail="You don't have permission to access this resource.",
            )
        request.scope["user"] = user_id
    except urllib.error.URLError as error:
        logger.error(f"Unable to contact server for token validation {jwks_url} Message: {error}")
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail="Unable to contact server for token validation",
        ) from error

    except jwt.InvalidAlgorithmError as error:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Algorithm is invalid for decoding token",
            headers={"WWW-Authenticate": "Bearer"},
        ) from error
    except jwt.PyJWKClientError as error:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Unable to find the signing key",
            headers={"WWW-Authenticate": "Bearer"},
        ) from error
    except jwt.InvalidKeyError as error:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="JWKS key is not in the proper format",
            headers={"WWW-Authenticate": "Bearer"},
        ) from error
    except jwt.InvalidIssuerError as error:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Token's issuer claim does not match with the expected issuer",
            headers={"WWW-Authenticate": "Bearer"},
        ) from error
    except jwt.ExpiredSignatureError as error:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Token has expired",
            headers={"WWW-Authenticate": "Bearer"},
        ) from error
    except jwt.InvalidSignatureError as error:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Token's signature did not match",
            headers={"WWW-Authenticate": "Bearer"},
        ) from error
    except jwt.DecodeError as error:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Token can not be decoded because it failed validation",
            headers={"WWW-Authenticate": "Bearer"},
        ) from error
    except jwt.InvalidTokenError as error:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Invalid token",
            headers={"WWW-Authenticate": "Bearer"},
        ) from error


async def add_security_headers(request: Request, call_next):
    """
    Function that is used to add several security headers to the API
    """
    security_headers = {
        "Strict-Transport-Security": STRICT_TRANSPORT_SECURITY,
        "Cache-Control": CACHE_CONTROL,
        "Cross-Origin-Resource-Policy": CROSS_ORIGIN_RESOURCE_POLICY,
        "Referrer-Policy": REFERRER_POLICY,
        "X-Permitted-Cross-Domain-Policies": X_PERMITTED_CROSS_DOMAIN_POLICIES,
        "X-Content-Type-Options": X_CONTENT_TYPE_OPTIONS,
        "X-Frame-Options": X_FRAME_OPTIONS,
        "X-XSS-Protection": X_XSS_PROTECTION,
        "Content-Security-Policy": CONTENT_SECURITY_POLICY,
    }
    response = await call_next(request)
    for header, value in security_headers.items():
        response.headers[header] = value
    return response


def user_is_authorized(claims: dict, env_variables: dict) -> bool:
    """
    Function that is used to determine if the user has the required claim and value to access the web service
    """
    claim_key = env_variables[SSO_JWT_CLAIM_KEY_AUTHORIZATION]
    claim_check_value = env_variables[SSO_JWT_CLAIM_VALUE_AUTHORIZATION]
    return bool(claim_key in claims and claim_check_value in claims[claim_key])


def requires_no_auth(request: Request):
    """
    Function that is used for unauthenticated access
    """
    request.scope["user"] = "Anonymous"


def get_db_connection():
    db_connection = Session(bind=engine)
    try:
        yield db_connection
    finally:
        db_connection.close()


@retry(wait=wait_exponential(multiplier=1, min=2, max=10), stop=stop_after_attempt(100))
def check_db_initialized():
    required_table_names = [
        DBfinding.__tablename__,
        DBrepository.__tablename__,
        DBrule.__tablename__,
        DBscan.__tablename__,
        DBscanFinding.__tablename__,
    ]
    try:
        # Create a sqlalchemy inspector, will cause an exception if the db connection can't be established
        inspector = sqlalchemy.inspect(engine)

        # Check existence of required tables
        not_found_tables = []
        for table_name in required_table_names:
            table_exists = inspector.has_table(table_name)
            if not table_exists:
                not_found_tables.append(table_name)

        if len(not_found_tables) > 0:
            raise RuntimeError(f"Unable to determine existence of required table(s) " f"{', '.join(not_found_tables)}")
    except Exception as ex:
        logger.error(f"Database is NOT connected or initialized | {ex} | Retrying...")
        raise



# Third Party
from sqlalchemy import Column, ForeignKey, Integer

# First Party
from resc_backend.db.model import Base


class DBscanFinding(Base):
    __tablename__ = "scan_finding"
    finding_id = Column(Integer, ForeignKey("finding.id"), primary_key=True)
    scan_id = Column(Integer, ForeignKey("scan.id"), primary_key=True)

    def __init__(self, finding_id: int, scan_id: int):
        self.finding_id = finding_id
        self.scan_id = scan_id



# Standard Library
import logging
import sys
from argparse import ArgumentParser, Namespace

# First Party
from resc_backend.helpers.environment_wrapper import validate_environment
from resc_backend.helpers.rabbitmq.configuration import (
    REQUIRED_ENV_VARS,
    RESC_RABBITMQ_SERVICE_HOST,
    RESC_RABBITMQ_SERVICE_PORT_MGMT,
)
from resc_backend.helpers.rabbitmq.rabbitmq_initialization import (
    create_queue_user_and_set_permission,
    wait_for_rabbitmq_server_to_up,
)

env_variables = validate_environment(REQUIRED_ENV_VARS)
logger = logging.getLogger(__name__)


def create_cli_argparser() -> ArgumentParser:
    rabbitmq_service_host = f"{env_variables[RESC_RABBITMQ_SERVICE_HOST]}"
    rabbitmq_service_port = f"{env_variables[RESC_RABBITMQ_SERVICE_PORT_MGMT]}"
    rabbitmq_api_base_url = "http://" + rabbitmq_service_host + ":" + rabbitmq_service_port

    parser: ArgumentParser = ArgumentParser()
    parser.add_argument("--rabbitmq-url", type=str, default=rabbitmq_api_base_url, help="RabbitMQ URL")

    return parser


def validate_cli_arguments(args: Namespace):
    valid_arguments = True
    if not args.rabbitmq_url:
        logger.error("RabbtMQ URL needs to be specified")
        valid_arguments = False
    if not valid_arguments:
        return False
    return args


def bootstrap_rabbitmq_users():
    """
    This function creates users in RabbitMQ server along with required permissions.
    """
    parser: ArgumentParser = create_cli_argparser()
    args: Namespace = parser.parse_args()
    args = validate_cli_arguments(args)
    if not args:
        logger.error("CLI arguments validation failed while bootstrapping rabbitmq users")
        sys.exit(-1)

    server_up = wait_for_rabbitmq_server_to_up(rabbitmq_api_base_url=args.rabbitmq_url)
    if not server_up:
        raise TimeoutError("Wait for rabbitmq server to up has been failed.")
    create_queue_user_and_set_permission(rabbitmq_api_base_url=args.rabbitmq_url)



# Standard Library
from typing import Dict

# Third Party
from pydantic import BaseModel


class AuditCountOverTime(BaseModel):
    time_period: str
    audit_by_auditor_count: Dict[str, int]
    total: int = 0



from __future__ import annotations

# Standard Library
import logging

# Third Party
from fastapi import FastAPI, Request, status
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import RequestValidationError
from fastapi.responses import JSONResponse
from sqlalchemy.exc import OperationalError

# First Party
from resc_backend.constants import ERROR_MESSAGE_500, ERROR_MESSAGE_503


def add_exception_handlers(app: FastAPI):
    @app.exception_handler(RequestValidationError)
    async def validation_exception_handler(request: Request, exc: RequestValidationError):
        log_warning(request, exc, "422 Unprocessable Entity")
        return JSONResponse(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            content=jsonable_encoder({"detail": exc.errors(), "body": exc.body}),
        )

    @app.exception_handler(ValueError)
    async def value_error_exception_handler(request: Request, exc: ValueError):
        log_warning(request, exc, "400 Bad Request")
        return JSONResponse(
            status_code=status.HTTP_400_BAD_REQUEST,
            content=jsonable_encoder({"detail": exc.errors(), "body": exc.body}),
        )

    @app.exception_handler(Exception)
    async def internal_server_error_exception_handler(request: Request, exc: Exception):
        log_error(request, exc, "500 Internal Server Error")
        return JSONResponse(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            content={"detail": ERROR_MESSAGE_500},
        )

    @app.exception_handler(OperationalError)
    async def service_unavailable_exception_handler(request: Request, exc: OperationalError):
        log_error(request, exc, "503 Service Unavailable")
        return JSONResponse(
            status_code=status.HTTP_503_SERVICE_UNAVAILABLE,
            content={"detail": ERROR_MESSAGE_503},
        )


def log_error(request: Request, exc: Exception, response_status: str):
    """
        Send error details to log file
    :param request:
        Request object for the endpoint
    :param exc:
        Exception thrown by the endpoint
    :param response_status:
        Response status from the endpoint
    """
    logging.error(f"{request.method} {request.url} {response_status}")
    if hasattr(exc, "body"):
        logging.error(f"request body: {exc.body}")
    logging.error(f"error: {exc}")


def log_warning(request: Request, exc: Exception, response_status: str):
    """
        Send warning details to log file
    :param request:
        Request object for the endpoint
    :param exc:
        Exception thrown by the endpoint
    :param response_status:
        Response status from the endpoint
    """
    exc_str = f"{exc}".replace("\n", " ").replace("   ", " ")
    logging.warning(f"{request.method} {request.url} {response_status}")
    if hasattr(exc, "body"):
        logging.warning(f"request body: {exc.body}")
    logging.warning(f"warning: {exc_str}")



# Standard Library
import logging
from datetime import datetime
from typing import List, Optional

# Third Party
from fastapi import APIRouter, Depends, Query, status
from fastapi_cache.decorator import cache

# First Party
from resc_backend.constants import (
    CACHE_NAMESPACE_FINDING,
    CACHE_NAMESPACE_RULE,
    ERROR_MESSAGE_500,
    ERROR_MESSAGE_503,
    REDIS_CACHE_EXPIRE,
    RULES_TAG,
    RWS_ROUTE_DETECTED_RULES,
    RWS_ROUTE_FINDING_STATUS_COUNT,
    RWS_ROUTE_RULES,
)
from resc_backend.db.connection import Session
from resc_backend.resc_web_service.crud import finding as finding_crud
from resc_backend.resc_web_service.dependencies import get_db_connection
from resc_backend.resc_web_service.schema.finding_status import FindingStatus
from resc_backend.resc_web_service.schema.rule_count_model import RuleFindingCountModel
from resc_backend.resc_web_service.schema.status_count import StatusCount
from resc_backend.resc_web_service.schema.vcs_provider import VCSProviders

router = APIRouter(tags=[RULES_TAG])

logger = logging.getLogger(__name__)


@router.get(
    f"{RWS_ROUTE_DETECTED_RULES}",
    response_model=List[str],
    summary="Get unique rules from findings",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve all the unique detected rules across all the findings"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
@cache(namespace=CACHE_NAMESPACE_RULE, expire=REDIS_CACHE_EXPIRE)
def get_distinct_rules_from_findings(
    finding_statuses: List[FindingStatus] = Query(None, alias="findingstatus", title="FindingStatuses"),
    vcs_providers: List[VCSProviders] = Query(None, alias="vcsprovider", title="VCSProviders"),
    project_name: Optional[str] = Query("", pattern=r"^[A-z0-9 .\-_%]*$"),
    repository_name: Optional[str] = Query("", pattern=r"^[A-z0-9 .\-_%]*$"),
    start_date_time: Optional[datetime] = Query(None),
    end_date_time: Optional[datetime] = Query(None),
    rule_pack_versions: Optional[List[str]] = Query(None, alias="rule_pack_version", title="RulePackVersion"),
    db_connection: Session = Depends(get_db_connection),
) -> List[str]:
    """
        Retrieve all uniquely detected rules across all findings in the database

    - **db_connection**: Session of the database connection
    - **finding_statuses**: Optional, filter on supported finding statuses
    - **vcs_providers**: Optional, filter on supported vcs provider types
    - **project_name**: Optional, filter on project name. It is used as a full string match filter
    - **repository_name**: Optional, filter on repository name. It is used as a string contains filter
    - **start_date_time**: Optional, filter on start date
    - **end_date_time**: Optional, filter on end date
    - **rule_pack_version**: Optional, filter on rule pack version
    - **return**: List[str] The output will contain a list of strings of unique rules in the findings table
    """
    distinct_rules = finding_crud.get_distinct_rules_from_findings(
        db_connection,
        finding_statuses=finding_statuses,
        vcs_providers=vcs_providers,
        project_name=project_name,
        repository_name=repository_name,
        start_date_time=start_date_time,
        end_date_time=end_date_time,
        rule_pack_versions=rule_pack_versions,
    )
    rules = [rule.rule_name for rule in distinct_rules]
    return rules


@router.get(
    f"{RWS_ROUTE_RULES}{RWS_ROUTE_FINDING_STATUS_COUNT}",
    response_model=List[RuleFindingCountModel],
    summary="Get detected rules with counts per status",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve all the detected rules with counts per status"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
@cache(namespace=CACHE_NAMESPACE_FINDING, expire=REDIS_CACHE_EXPIRE)
def get_rules_finding_status_count(
    rule_pack_versions: Optional[List[str]] = Query(None, alias="rule_pack_version", title="RulePackVersion"),
    rule_tags: Optional[List[str]] = Query(None, alias="rule_tag", title="RuleTag"),
    db_connection: Session = Depends(get_db_connection),
) -> List[RuleFindingCountModel]:
    """
        Retrieve all detected rules with finding counts per supported status

    - **rule_pack_version**: Optional, filter on rule pack version
    - **rule_tag**: Optional, filter on rule tag
    - **db_connection**: Session of the database connection
    - **return**: List[str] The output will contain a list of strings of unique rules with counts per status
    """
    rule_finding_counts = finding_crud.get_rule_findings_count_by_status(
        db_connection, rule_pack_versions=rule_pack_versions, rule_tags=rule_tags
    )
    rule_findings_counts = []

    for rule_name, rule_counts in rule_finding_counts.items():
        rule_finding_count = RuleFindingCountModel(
            rule_name=rule_name, finding_count=rule_counts["total_findings_count"]
        )
        rule_finding_count.finding_statuses_count.append(
            StatusCount(status=FindingStatus.TRUE_POSITIVE.value, count=rule_counts["true_positive"])
        )
        rule_finding_count.finding_statuses_count.append(
            StatusCount(status=FindingStatus.FALSE_POSITIVE.value, count=rule_counts["false_positive"])
        )
        rule_finding_count.finding_statuses_count.append(
            StatusCount(status=FindingStatus.NOT_ANALYZED.value, count=rule_counts["not_analyzed"])
        )
        rule_finding_count.finding_statuses_count.append(
            StatusCount(status=FindingStatus.UNDER_REVIEW.value, count=rule_counts["under_review"])
        )
        rule_finding_count.finding_statuses_count.append(
            StatusCount(
                status=FindingStatus.CLARIFICATION_REQUIRED.value,
                count=rule_counts["clarification_required"],
            )
        )
        rule_findings_counts.append(rule_finding_count)

    return rule_findings_counts



# Standard Library
from enum import Enum


class FindingStatus(str, Enum):
    NOT_ANALYZED = "NOT_ANALYZED"
    UNDER_REVIEW = "UNDER_REVIEW"
    CLARIFICATION_REQUIRED = "CLARIFICATION_REQUIRED"
    FALSE_POSITIVE = "FALSE_POSITIVE"
    TRUE_POSITIVE = "TRUE_POSITIVE"



# Standard Library
import logging
from typing import Optional

# Third Party
import requests
from requests import Response

# First Party
from resc_backend.constants import (
    DEFAULT_RECORDS_PER_PAGE_LIMIT,
    RWS_ROUTE_RULE_PACKS,
    RWS_VERSION_PREFIX,
)

logger = logging.getLogger(__name__)


def upload_rule_pack_toml_file(url: str, rule_file_path: str):
    with open(rule_file_path, "rb") as toml_file:
        files = {
            "rule_file": (
                "RESC-SECRETS-RULE.toml",
                toml_file,
                "application/octet-stream",
            )
        }
        response = requests.post(url=url, files=files, proxies={"http": "", "https": ""}, timeout=10)
        toml_file.close()
    return response


def download_rule_pack_toml_file(rws_url: str, rule_pack_version: Optional[str] = "") -> Response:
    params = {}
    if rule_pack_version:
        params = {"rule_pack_version": rule_pack_version}
    response = requests.get(
        url=f"{rws_url}{RWS_VERSION_PREFIX}{RWS_ROUTE_RULE_PACKS}",
        params=params,
        timeout=10,
    )

    if response.status_code == 200:
        logger.debug(f"Rule pack version: {rule_pack_version} has been successfully downloaded")
    else:
        logger.error(
            f"Downloading rule pack version {rule_pack_version} failed with "
            f"error: {response.status_code}->{response.text}"
        )
    return response


def get_rule_packs(
    url: str,
    version: Optional[str] = None,
    active: Optional[bool] = None,
    skip: Optional[int] = 0,
    limit: Optional[int] = DEFAULT_RECORDS_PER_PAGE_LIMIT,
):
    api_url = f"{url}{RWS_VERSION_PREFIX}{RWS_ROUTE_RULE_PACKS}/versions"
    params = {"active": active, "skip": skip, "limit": limit}
    if version:
        params["version"] = version
    response = requests.get(api_url, params=params, proxies={"http": "", "https": ""}, timeout=10)
    return response



# Standard Library
from datetime import datetime
from typing import List

# Third Party
from pydantic import validator
from pydantic.dataclasses import dataclass

# First Party
from resc_backend.resc_web_service.schema.finding_status import FindingStatus
from resc_backend.resc_web_service.schema.vcs_provider import VCSProviders


@dataclass
class FindingsFilter:
    vcs_providers: List[VCSProviders] = None
    finding_statuses: List[FindingStatus] = None
    rule_names: List[str] = None
    rule_tags: List[str] = None
    project_name: str = None
    repository_name: str = None
    scan_ids: List[int] = None
    start_date_time: datetime = None
    end_date_time: datetime = None
    event_sent: bool = None
    rule_pack_versions: List[str] = None

    @validator("end_date_time")
    @classmethod
    def date_range_check(cls, end_date_time: datetime, values: dict):
        if end_date_time and values["start_date_time"]:
            if values["start_date_time"] >= end_date_time:
                raise ValueError("the start of the date range needs to be prior to the end of it.")

        return end_date_time



# Standard Library
import logging
from typing import Any, Callable, Dict, Tuple

# Third Party
from fastapi import Request, Response
from fastapi_cache import FastAPICache
from fastapi_cache.backends.redis import RedisBackend
from redis import asyncio as aioredis

# First Party
from resc_backend.common import initialise_logs
from resc_backend.constants import CACHE_PREFIX, LOG_FILE_CACHING
from resc_backend.helpers.environment_wrapper import validate_environment
from resc_backend.resc_web_service.configuration import (
    CONDITIONAL_REDIS_ENV_VARS,
    REDIS_PASSWORD,
    RESC_REDIS_CACHE_ENABLE,
    RESC_REDIS_SERVICE_HOST,
    RESC_REDIS_SERVICE_PORT,
)

logger_config = initialise_logs(LOG_FILE_CACHING)
logger = logging.getLogger(__name__)


class CacheManager:
    @classmethod
    def initialize_cache(cls, env_variables):
        cache_enabled = env_variables[RESC_REDIS_CACHE_ENABLE].lower() in ["true"]
        if cache_enabled:
            env_variables.update(validate_environment(CONDITIONAL_REDIS_ENV_VARS))
            redis_host = f"{env_variables[RESC_REDIS_SERVICE_HOST]}"
            redis_port = f"{env_variables[RESC_REDIS_SERVICE_PORT]}"
            redis_password = f"{env_variables[REDIS_PASSWORD]}"
            redis_backend = cls.get_cache_client(host=redis_host, port=int(redis_port), password=redis_password)
            FastAPICache.init(
                RedisBackend(redis_backend),
                prefix=CACHE_PREFIX,
                key_builder=cls.request_key_builder,
                enable=cache_enabled,
            )
        else:
            FastAPICache.init(backend=RedisBackend(None), enable=cache_enabled)

    @staticmethod
    def get_cache_client(host: str, port: int, password: str):
        cache_client = aioredis.from_url(f"redis://{host}:{port}", password=password)
        return cache_client

    @staticmethod
    def request_key_builder(
        func: Callable[..., Any],
        namespace: str = "",
        *,
        request: Request = None,
        response: Response = None,
        args: Tuple[Any, ...],
        kwargs: Dict[str, Any],
    ) -> str:
        """
        Build a unique key for caching based on the provided function, namespace, request, response,
        arguments (args), and keyword arguments (kwargs).

        Args:
            func (Callable): The function for which the key is being built.
            namespace (str, optional): A namespace to differentiate keys (default "").
            request (Request, optional): The request object (default None).
            response (Response, optional): The response object (default None).
            args (Tuple[Any, ...]): Positional arguments passed to the function.
            kwargs (Dict[str, Any]): Keyword arguments passed to the function.

        Returns:
            str: A unique key based on the input parameters.
        """

        cache_key = ":".join(
            [
                FastAPICache.get_prefix(),
                namespace,
                request.method.lower(),
                request.url.path,
                repr(sorted(request.query_params.items())),
            ]
        )
        logger.debug(f"Cache created with key: {cache_key}")
        return cache_key

    @staticmethod
    def personalized_key_builder(
        func: Callable[..., Any],
        namespace: str = "",
        *,
        request: Request = None,
        response: Response = None,
        args: Tuple[Any, ...],
        kwargs: Dict[str, Any],
    ) -> str:
        """
        Build a personalized unique key for caching based logged-in user on the provided function, namespace, request,
        response, arguments (args), and keyword arguments (kwargs).
        This should be used for apis specific to logged-in user.

        Args:
            func (Callable): The function for which the key is being built.
            namespace (str, optional): A namespace to differentiate keys (default "").
            request (Request, optional): The request object (default None).
            response (Response, optional): The response object (default None).
            args (Tuple[Any, ...]): Positional arguments passed to the function.
            kwargs (Dict[str, Any]): Keyword arguments passed to the function.

        Returns:
            str: A unique key based on the input parameters.
        """
        cache_key = ":".join(
            [
                CACHE_PREFIX,
                namespace,
                request.user,
                request.method.lower(),
                request.url.path,
                repr(sorted(request.query_params.items())),
            ]
        )
        logger.debug(f"Cache created with key: {cache_key}")
        return cache_key

    @staticmethod
    async def clear_cache_by_namespace(namespace):
        cache_enabled = FastAPICache.get_enable()
        if cache_enabled:
            await FastAPICache.clear(namespace=namespace)
            logger.debug(f"Cache cleared for namespaces: {namespace}")

    @staticmethod
    async def clear_all_cache():
        cache_enabled = FastAPICache.get_enable()
        if cache_enabled:
            await FastAPICache.clear()
            logger.debug("Cache cleared for all namespaces")



# Standard Library
from typing import List

# Third Party
from sqlalchemy import func
from sqlalchemy.orm import Session
from sqlalchemy.orm.query import Query

# First Party
from resc_backend.constants import (
    DEFAULT_RECORDS_PER_PAGE_LIMIT,
    MAX_RECORDS_PER_PAGE_LIMIT,
)
from resc_backend.db.model import (
    DBaudit,
    DBrepository,
    DBscan,
    DBscanFinding,
    DBVcsInstance,
)
from resc_backend.resc_web_service.crud import finding as finding_crud
from resc_backend.resc_web_service.crud import scan as scan_crud
from resc_backend.resc_web_service.crud import scan_finding as scan_finding_crud
from resc_backend.resc_web_service.schema import repository as repository_schema
from resc_backend.resc_web_service.schema.finding_status import FindingStatus
from resc_backend.resc_web_service.schema.scan_type import ScanType
from resc_backend.resc_web_service.schema.vcs_provider import VCSProviders


def _get_max_base_scan(db_connection: Session) -> Query:
    subquery: Query = db_connection.query(DBscan.repository_id, func.max(DBscan.id_).label("latest_base_scan_id"))
    subquery = subquery.where(DBscan.scan_type == ScanType.BASE)
    subquery = subquery.where(DBscan.is_latest == True)  # noqa: E712
    subquery = subquery.group_by(DBscan.repository_id)
    return subquery.subquery()


def get_repositories(
    db_connection: Session,
    vcs_providers: List[VCSProviders] = None,
    skip: int = 0,
    limit: int = DEFAULT_RECORDS_PER_PAGE_LIMIT,
    project_filter: str = "",
    repository_filter: str = "",
    only_if_has_findings: bool = False,
):
    """
        Retrieve repository records optionally filtered
    :param db_connection:
        Session of the database connection
    :param skip:
        integer amount of records to skip to support pagination
    :param limit:
        integer amount of records to return, to support pagination
    :param vcs_providers:
        optional [string] filtering the VCS provider
    :param project_filter:
        optional, filter on project name. Is used as a string contains filter
    :param repository_filter:
        optional, filter on repository name. Is used as a string contains filter
    :param only_if_has_findings:
        optional, filter on repositories with findings
    :return: repositories
        list of DBrepository objects
    """
    limit_val = MAX_RECORDS_PER_PAGE_LIMIT if limit > MAX_RECORDS_PER_PAGE_LIMIT else limit

    # Get the latest scan for repository
    sub_query: Query = db_connection.query(DBscan.repository_id, func.max(DBscan.timestamp).label("max_timestamp"))
    sub_query = sub_query.where(DBscan.is_latest == True)  # noqa: E712
    sub_query = sub_query.group_by(DBscan.repository_id)
    sub_query = sub_query.subquery()

    query = db_connection.query(
        DBrepository.id_,
        DBrepository.project_key,
        DBrepository.repository_id,
        DBrepository.repository_name,
        DBrepository.repository_url,
        DBrepository.vcs_instance,
        DBVcsInstance.provider_type,
        func.coalesce(DBscan.id_, None).label("last_scan_id"),
        func.coalesce(DBscan.timestamp, None).label("last_scan_timestamp"),
    )
    query = query.join(DBVcsInstance, DBVcsInstance.id_ == DBrepository.vcs_instance)
    query = query.join(
        sub_query,
        DBrepository.id_ == sub_query.c.repository_id,
        isouter=True,
    )
    query = query.join(
        DBscan,
        (DBscan.repository_id == sub_query.c.repository_id) & (DBscan.timestamp == sub_query.c.max_timestamp),
        isouter=True,
    )

    if only_if_has_findings:
        last_scan_sub_query = _get_max_base_scan(db_connection)

        sub_query = db_connection.query(DBrepository.id_)
        sub_query = sub_query.join(
            last_scan_sub_query,
            DBrepository.id_ == last_scan_sub_query.c.repository_id,
        )
        sub_query = sub_query.join(
            DBscan,
            DBrepository.id_ == DBscan.repository_id,
        )
        sub_query = sub_query.where(DBscan.id_ >= last_scan_sub_query.c.latest_base_scan_id)
        sub_query = sub_query.join(DBscanFinding, DBscan.id_ == DBscanFinding.scan_id)
        sub_query = sub_query.distinct()

        # Filter on repositories that are in the selection
        query = query.where(DBrepository.id_.in_(sub_query))

    if vcs_providers and vcs_providers is not None:
        query = query.where(DBVcsInstance.provider_type.in_(vcs_providers))

    if project_filter:
        query = query.where(DBrepository.project_key == project_filter)

    if repository_filter:
        query = query.where(DBrepository.repository_name == repository_filter)

    repositories = query.order_by(DBrepository.id_).offset(skip).limit(limit_val).all()

    return repositories


def get_repositories_count(
    db_connection: Session,
    vcs_providers: List[VCSProviders] = None,
    project_filter: str = "",
    repository_filter: str = "",
    only_if_has_findings: bool = False,
) -> int:
    """
        Retrieve count of repository records optionally filtered
    :param db_connection:
        Session of the database connection
    :param vcs_providers:
        optional [string] filtering the VCS provider
    :param project_filter:
        optional, filter on project name
    :param repository_filter:
        optional, filter on repository name
    :param only_if_has_findings:
        optional, filter on repositories with findings
    :return: total_count
        count of repositories
    """
    query = db_connection.query(func.count(DBrepository.id_))

    if only_if_has_findings:
        last_scan_sub_query = _get_max_base_scan(db_connection)

        sub_query = db_connection.query(DBrepository.id_)
        sub_query = sub_query.join(
            last_scan_sub_query,
            DBrepository.id_ == last_scan_sub_query.c.repository_id,
        )
        sub_query = sub_query.join(
            DBscan,
            DBrepository.id_ == DBscan.repository_id,
        )
        sub_query = sub_query.where(DBscan.id_ >= last_scan_sub_query.c.latest_base_scan_id)
        sub_query = sub_query.join(DBscanFinding, DBscan.id_ == DBscanFinding.scan_id)
        sub_query = sub_query.distinct()

        # Filter on repositories that are in the selection
        query = query.where(DBrepository.id_.in_(sub_query))

    if vcs_providers and vcs_providers is not None:
        query = query.join(DBVcsInstance, DBVcsInstance.id_ == DBrepository.vcs_instance)
        query = query.where(DBVcsInstance.provider_type.in_(vcs_providers))

    if project_filter:
        query = query.where(DBrepository.project_key == project_filter)

    if repository_filter:
        query = query.where(DBrepository.repository_name == repository_filter)

    total_count = query.scalar()
    return total_count


def get_repository(db_connection: Session, repository_id: int):
    repository = db_connection.query(DBrepository).where(DBrepository.id_ == repository_id).first()
    return repository


def update_repository(
    db_connection: Session,
    repository_id: int,
    repository: repository_schema.RepositoryCreate,
):
    db_repository = db_connection.query(DBrepository).filter_by(id_=repository_id).first()

    db_repository.repository_name = repository.repository_name
    db_repository.repository_url = repository.repository_url
    db_repository.vcs_instance = repository.vcs_instance

    db_connection.commit()
    db_connection.refresh(db_repository)
    return db_repository


def create_repository(db_connection: Session, repository: repository_schema.RepositoryCreate):
    db_repository = DBrepository(
        project_key=repository.project_key,
        repository_id=repository.repository_id,
        repository_name=repository.repository_name,
        repository_url=repository.repository_url,
        vcs_instance=repository.vcs_instance,
    )
    db_connection.add(db_repository)
    db_connection.commit()
    db_connection.refresh(db_repository)
    return db_repository


def create_repository_if_not_exists(db_connection: Session, repository: repository_schema.RepositoryCreate):
    # Query the database to see if the repository object exists based on the unique constraint parameters
    query: Query = db_connection.query(DBrepository)
    query = query.where(DBrepository.project_key == repository.project_key)
    query = query.where(DBrepository.repository_id == repository.repository_id)
    query = query.where(DBrepository.vcs_instance == repository.vcs_instance)
    db_select_repository = query.first()

    if db_select_repository is not None:
        return db_select_repository

    # Create non-existing repository object
    return create_repository(db_connection, repository)


def get_distinct_projects(
    db_connection: Session,
    vcs_providers: List[VCSProviders] = None,
    repository_filter: str = "",
    only_if_has_findings: bool = False,
):
    """
        Retrieve all unique project names
    :param db_connection:
        Session of the database connection
    :param vcs_providers:
        optional, filter of supported vcs provider types
    :param repository_filter:
        optional, filter on repository name. Is used as a string contains filter
    :param only_if_has_findings:
        optional, filter on repositories that have findings
    :return: distinct_projects
        The output will contain a list of unique projects
    """
    query = db_connection.query(DBrepository.project_key)

    if only_if_has_findings:
        last_scan_sub_query = _get_max_base_scan(db_connection)
        query = query.join(
            last_scan_sub_query,
            DBrepository.id_ == last_scan_sub_query.c.repository_id,
        )
        query = query.join(
            DBscan,
            DBrepository.id_ == DBscan.repository_id,
        )
        query = query.where(DBscan.id_ >= last_scan_sub_query.c.latest_base_scan_id)
        query = query.join(DBscanFinding, DBscan.id_ == DBscanFinding.scan_id)

    if vcs_providers and vcs_providers is not None:
        query = query.join(DBVcsInstance, DBVcsInstance.id_ == DBrepository.vcs_instance)
        query = query.where(DBVcsInstance.provider_type.in_(vcs_providers))

    if repository_filter:
        query = query.where(DBrepository.repository_name == repository_filter)

    query = query.distinct()
    distinct_projects = query.all()
    return distinct_projects


def get_distinct_repositories(
    db_connection: Session,
    vcs_providers: List[VCSProviders] = None,
    project_name: str = "",
    only_if_has_findings: bool = False,
):
    """
        Retrieve all unique repository names
    :param db_connection:
        Session of the database connection
    :param vcs_providers:
        optional, filter of supported vcs provider types
    :param project_name:
        optional, filter on project name. Is used as a full string match filter
    :param only_if_has_findings:
        optional, filter on repositories that have findings
    :return: distinct_repositories
        The output will contain a list of unique repositories
    """
    query = db_connection.query(DBrepository.repository_name)

    if only_if_has_findings:
        last_scan_sub_query = _get_max_base_scan(db_connection)
        query = query.join(
            last_scan_sub_query,
            DBrepository.id_ == last_scan_sub_query.c.repository_id,
        )
        query = query.join(
            DBscan,
            DBrepository.id_ == DBscan.repository_id,
        )
        query = query.where(DBscan.id_ >= last_scan_sub_query.c.latest_base_scan_id)
        query = query.join(DBscanFinding, DBscan.id_ == DBscanFinding.scan_id)

    if vcs_providers and vcs_providers is not None:
        query = query.join(DBVcsInstance, DBVcsInstance.id_ == DBrepository.vcs_instance)
        query = query.where(DBVcsInstance.provider_type.in_(vcs_providers))

    if project_name:
        query = query.where(DBrepository.project_key == project_name)

    query = query.distinct()
    distinct_repositories = query.all()
    return distinct_repositories


def get_findings_metadata_by_repository_id(db_connection: Session, repository_ids: list[int]):
    """
        Retrieves the finding metadata for a repository id from the database with most recent scan information
    :param db_connection:
        Session of the database connection
    :param repository_ids:
        ids of the repository for which findings metadata to be retrieved
    :return: findings_metadata
        findings_metadata containing the count for each status
    """
    query = db_connection.query(DBrepository.id_, DBaudit.status, func.count(DBscanFinding.finding_id))

    last_scan_sub_query = _get_max_base_scan(db_connection)
    query = query.join(
        last_scan_sub_query,
        DBrepository.id_ == last_scan_sub_query.c.repository_id,
    )
    query = query.join(
        DBscan,
        DBrepository.id_ == DBscan.repository_id,
    )
    query = query.where(DBscan.id_ >= last_scan_sub_query.c.latest_base_scan_id)
    query = query.join(DBscanFinding, DBscan.id_ == DBscanFinding.scan_id)
    query = query.join(
        DBaudit,
        (DBaudit.finding_id == DBscanFinding.finding_id) & (DBaudit.is_latest == True),  # noqa: E712
        isouter=True,
    )
    query = query.where(DBrepository.id_.in_(repository_ids))
    query = query.group_by(
        DBrepository.id_,
        DBaudit.status,
    )
    status_counts = query.all()
    repo_count_dict = {}
    for repository_id in repository_ids:
        repo_count_dict[repository_id] = {
            "true_positive": 0,
            "false_positive": 0,
            "not_analyzed": 0,
            "under_review": 0,
            "clarification_required": 0,
            "total_findings_count": 0,
        }
    for status_count in status_counts:
        repo_count_dict[status_count[0]]["total_findings_count"] += status_count[2]
        if status_count[1] == FindingStatus.NOT_ANALYZED.value or status_count[1] is None:
            repo_count_dict[status_count[0]]["not_analyzed"] += status_count[2]
        elif status_count[1] == FindingStatus.FALSE_POSITIVE.value:
            repo_count_dict[status_count[0]]["false_positive"] += status_count[2]
        elif status_count[1] == FindingStatus.TRUE_POSITIVE.value:
            repo_count_dict[status_count[0]]["true_positive"] += status_count[2]
        elif status_count[1] == FindingStatus.UNDER_REVIEW.value:
            repo_count_dict[status_count[0]]["under_review"] += status_count[2]
        elif status_count[1] == FindingStatus.CLARIFICATION_REQUIRED.value:
            repo_count_dict[status_count[0]]["clarification_required"] += status_count[2]

    return repo_count_dict


def delete_repository(db_connection: Session, repository_id: int, delete_related: bool = False):
    """
        Delete a repository object
    :param db_connection:
        Session of the database connection
    :param repository_id:
        id of the repository to be deleted
    :param delete_related:
        if related records need to be deleted
    """
    if delete_related:
        scan_finding_crud.delete_scan_finding_by_repository_id(db_connection, repository_id=repository_id)
        finding_crud.delete_findings_by_repository_id(db_connection, repository_id=repository_id)
        scan_crud.delete_scans_by_repository_id(db_connection, repository_id=repository_id)
    db_connection.query(DBrepository).where(DBrepository.id_ == repository_id).delete(synchronize_session=False)
    db_connection.commit()


def delete_repositories_by_vcs_instance_id(db_connection: Session, vcs_instance_id: int):
    """
        Delete repositories for a given vcs instance
    :param db_connection:
        Session of the database connection
    :param vcs_instance_id:
        id of the vcs instance
    """
    db_connection.query(DBrepository).where(
        DBrepository.vcs_instance == DBVcsInstance.id_,
        DBVcsInstance.id_ == vcs_instance_id,
    ).delete(synchronize_session=False)
    db_connection.commit()



# Standard Library
import logging
from typing import List, Optional

# Third Party
from packaging.version import Version
from sqlalchemy import func, update
from sqlalchemy.orm import Session
from sqlalchemy.orm.query import Query
from sqlalchemy.sql.expression import true

# First Party
from resc_backend.constants import (
    DEFAULT_RECORDS_PER_PAGE_LIMIT,
    MAX_RECORDS_PER_PAGE_LIMIT,
)
from resc_backend.db.model import DBrule, DBrulePack, DBruleTag, DBtag
from resc_backend.resc_web_service.schema import rule_pack as rule_pack_schema

logger = logging.getLogger(__name__)


def get_rule_pack(db_connection: Session, version: Optional[str]) -> rule_pack_schema.RulePackRead:
    """
        Get active rule pack from database
    :param db_connection:
        Session of the database connection
    :param version:
        optional, version of the rule pack to be fetched else latest rule pack version will be fetched
    :return: RulePackRead
        The output returns RulePackRead type object
    """
    query = db_connection.query(DBrulePack)
    if version:
        query = query.where(DBrulePack.version == version)
    else:
        logger.debug("rule pack version not specified, fetching currently active one")
        query = query.where(DBrulePack.active == true())
    rule_pack = query.first()
    return rule_pack


def create_rule_pack_version(db_connection: Session, rule_pack: rule_pack_schema.RulePackCreate):
    """
        Create rule pack version in database
    :param db_connection:
        Session of the database connection
    :param rule_pack:
        RulePackCreate object to be created
    """
    db_rule_pack = DBrulePack(
        version=rule_pack.version,
        global_allow_list=rule_pack.global_allow_list,
        active=rule_pack.active,
    )
    db_connection.add(db_rule_pack)
    db_connection.commit()
    db_connection.refresh(db_rule_pack)
    return db_rule_pack


def get_newest_rule_pack(db_connection: Session) -> rule_pack_schema.RulePackRead:
    """
        Fetch the newest rule pack from database
    :param db_connection:
        Session of the database connection
    :return: RulePackRead
        The output returns RulePackRead type object
    """
    rule_packs = db_connection.query(DBrulePack).all()
    newest_rule_pack = None
    if rule_packs:
        newest_rule_pack: rule_pack_schema.RulePackRead = rule_packs[0]
        for rule_pack in rule_packs[1:]:
            if Version(rule_pack.version) > Version(newest_rule_pack.version):
                newest_rule_pack = rule_pack
    return newest_rule_pack


def get_rule_packs(
    db_connection: Session,
    version: str = None,
    active: bool = None,
    skip: int = 0,
    limit: int = DEFAULT_RECORDS_PER_PAGE_LIMIT,
) -> List[DBrulePack]:
    """
        Retrieve rule packs from database
    :param db_connection:
        Session of the database connection
    :param version:
        optional, filter on rule pack version
    :param active:
        optional, filter on active rule pack
    :param skip:
        integer amount of records to skip, to support pagination
    :param limit:
        integer amount of records to return, to support pagination
    :return: [RulePackRead]
        The output will contain a PaginationModel containing the list of RulePackRead type objects,
        or an empty list if no rule pack was found
    """
    limit_val = MAX_RECORDS_PER_PAGE_LIMIT if limit > MAX_RECORDS_PER_PAGE_LIMIT else limit
    query = db_connection.query(DBrulePack)

    if version:
        query = query.where(DBrulePack.version == version)
    if active is not None:
        query = query.where(DBrulePack.active == active)
    rule_packs = query.order_by(DBrulePack.version.desc()).offset(skip).limit(limit_val).all()
    return rule_packs


def get_current_active_rule_pack(
    db_connection: Session,
) -> Optional[DBrulePack]:
    """
        Return the currently active rule_pack, if any.
    :param db_connection:
        Session of the database connection
    :return: DBRulePack
        returns the DBRulePack containing the active rule pack
    """
    query = db_connection.query(DBrulePack)
    active_rule_pack = query.where(DBrulePack.active == 1).one()
    return active_rule_pack


def get_rule_packs_tags(db_connection: Session, versions: list) -> List[str]:
    """
        Retrieve rule packs tags for versions from database
    :param db_connection:
        Session of the database connection
    :param versions:
        optional, filter on rule pack version
    :return: [str]
        The output will contain the list of str that are the tags, or an empty list.
    """

    query: Query = db_connection.query(DBtag.name)
    query = query.join(DBruleTag, DBruleTag.tag_id == DBtag.id_)
    query = query.join(DBrule, DBrule.id_ == DBruleTag.rule_id)
    query = query.where(DBrule.rule_pack.in_(versions))
    rule_packs_tags = query.distinct().all()
    rule_packs_tags = [t for (t,) in rule_packs_tags]
    return rule_packs_tags


def get_total_rule_packs_count(db_connection: Session, version: str = None, active: bool = None) -> int:
    """
        Retrieve total count of rule packs from database
    :param db_connection:
        Session of the database connection
    :param version:
        optional, filter on rule pack version
    :param active:
        optional, filter on active rule pack
    :return: int
        The output contains total count of rule packs
    """
    total_count_query = db_connection.query(func.count(DBrulePack.version))
    if version:
        total_count_query = total_count_query.where(DBrulePack.version == version)
    if active is not None:
        total_count_query = total_count_query.where(DBrulePack.active == active)

    total_count = total_count_query.scalar()
    return total_count


def make_older_rule_packs_to_inactive(latest_rule_pack_version: str, db_connection: Session):
    """
        Make older rule packs to inactive
    :param latest_rule_pack_version:
        latest rule pack version
    :param db_connection:
        Session of the database connection
    """
    db_connection.execute(update(DBrulePack).where(DBrulePack.version != latest_rule_pack_version).values(active=False))
    db_connection.commit()



# Standard Library
import html
from datetime import datetime

# Third Party
from sqlalchemy import (
    Boolean,
    Column,
    DateTime,
    Enum,
    ForeignKey,
    Integer,
    String,
    text,
)

# First Party
from resc_backend.db.model import Base
from resc_backend.resc_web_service.schema.finding_status import FindingStatus


class DBaudit(Base):
    __tablename__ = "audit"
    id_ = Column("id", Integer, primary_key=True)
    finding_id = Column(Integer, ForeignKey("finding.id"), nullable=False)
    status = Column(
        Enum(FindingStatus),
        default=FindingStatus.NOT_ANALYZED.value,
        server_default=text("NOT_ANALYZED"),
        nullable=False,
    )
    auditor = Column(String(200))
    comment = Column(String(255), nullable=True)
    timestamp = Column(DateTime, nullable=False, default=datetime.utcnow)
    is_latest = Column(Boolean, nullable=False, default=False, server_default=text("0"))

    def __init__(self, finding_id, status, auditor, comment, timestamp, is_latest):
        sanitized_comment = html.escape(comment) if comment else comment
        self.finding_id = finding_id
        self.status = status
        self.auditor = auditor
        self.comment = sanitized_comment
        self.timestamp = timestamp
        self.is_latest = is_latest

    @staticmethod
    def create_from_metadata(
        finding_id: int,
        status: str,
        auditor: str,
        comment: str,
        timestamp: datetime,
        is_latest: bool,
    ):
        sanitized_comment = html.escape(comment) if comment else comment
        db_audit = DBaudit(
            finding_id=finding_id,
            auditor=auditor,
            status=status,
            comment=sanitized_comment,
            timestamp=timestamp,
            is_latest=is_latest,
        )
        return db_audit



# Standard Library
import logging

# Third Party
import requests

# First Party
from resc_backend.constants import RWS_ROUTE_VCS, RWS_VERSION_PREFIX
from resc_backend.resc_web_service.schema.vcs_instance import VCSInstanceCreate

logger = logging.getLogger(__name__)


def create_vcs_instance(url: str, vcs_instance: VCSInstanceCreate):
    api_url = f"{url}{RWS_VERSION_PREFIX}{RWS_ROUTE_VCS}"
    response = requests.post(api_url, data=vcs_instance.json(), proxies={"http": "", "https": ""}, timeout=10)
    return response



# Standard Library
from typing import List

# Third Party
from sqlalchemy.orm import Session

# First Party
from resc_backend.db.model import (
    DBfinding,
    DBrepository,
    DBscan,
    DBscanFinding,
    DBVcsInstance,
)


def create_scan_findings(db_connection: Session, scan_findings: List[DBscanFinding]) -> int:
    if len(scan_findings) < 1:
        # Function is called with an empty list of findings
        return 0

    # load existing scan findings for this scan into the session
    scan_id = scan_findings[0].scan_id
    _ = db_connection.query(DBscanFinding).where(DBscanFinding.scan_id == scan_id).all()

    # merge the new scan findings into the session, ignoring duplicates
    for scan_finding in scan_findings:
        db_connection.merge(scan_finding)

    db_connection.commit()

    return len(scan_findings)


def get_scan_findings(db_connection: Session, finding_id: int) -> List[DBscanFinding]:
    scan_findings = db_connection.query(DBscanFinding)
    scan_findings = scan_findings.where(DBscanFinding.finding_id == finding_id).all()
    return scan_findings


def delete_scan_finding(db_connection: Session, finding_id: int = None, scan_id: int = None):
    """
        Delete scan findings when finding id or scan id provided
    :param db_connection:
        Session of the database connection
    :param finding_id:
        optional, id of the finding
    :param scan_id:
        optional, id of the scan
    """
    if finding_id or scan_id:
        query = db_connection.query(DBscanFinding)
        if finding_id:
            query = query.where(DBscanFinding.finding_id == finding_id)
        if scan_id:
            query = query.where(DBscanFinding.scan_id == scan_id)
        query.delete(synchronize_session=False)
        db_connection.commit()


def delete_scan_finding_by_repository_id(db_connection: Session, repository_id: int):
    """
        Delete scan findings for a given repository
    :param db_connection:
        Session of the database connection
    :param repository_id:
        id of the repository
    """
    db_connection.query(DBscanFinding).where(
        DBscanFinding.scan_id == DBscan.id_,
        DBscanFinding.finding_id == DBfinding.id_,
        DBscan.repository_id == DBfinding.repository_id,
        DBscan.repository_id == repository_id,
    ).delete(synchronize_session=False)
    db_connection.commit()


def delete_scan_finding_by_vcs_instance_id(db_connection: Session, vcs_instance_id: int):
    """
        Delete scan findings for a given vcs instance
    :param db_connection:
        Session of the database connection
    :param vcs_instance_id:
        id of the vcs instance
    """
    db_connection.query(DBscanFinding).where(
        DBscanFinding.scan_id == DBscan.id_,
        DBscanFinding.finding_id == DBfinding.id_,
        DBscan.repository_id == DBrepository.id_,
        DBrepository.vcs_instance == DBVcsInstance.id_,
        DBVcsInstance.id_ == vcs_instance_id,
    ).delete(synchronize_session=False)
    db_connection.commit()



# Standard Library
from typing import Optional

# Third Party
from pydantic import BaseModel, conint, constr


class RuleAllowListBase(BaseModel):
    description: Optional[constr(max_length=2000)]
    regexes: Optional[str]
    paths: Optional[str]
    commits: Optional[str]
    stop_words: Optional[str]


class RuleAllowListCreate(RuleAllowListBase):
    @classmethod
    def create_from_base_class(cls, base_object: RuleAllowListBase):
        return cls(**(dict(base_object)))


class RuleAllowList(RuleAllowListBase):
    pass


class RuleAllowListRead(RuleAllowListBase):
    id_: conint(gt=0)

    class Config:
        orm_mode = True



# Standard Library
from typing import Optional

# Third Party
from pydantic import BaseModel, conint, conlist, constr, validator

# First Party
from resc_backend.constants import AZURE_DEVOPS
from resc_backend.resc_web_service.schema.vcs_provider import VCSProviders


class VCSInstanceBase(BaseModel):
    name: constr(max_length=200)
    provider_type: VCSProviders
    hostname: constr(max_length=200)
    port: conint(gt=-0, lt=65536)
    scheme: constr(max_length=20)
    exceptions: Optional[conlist(item_type=str, min_items=None, max_items=500)]
    scope: Optional[conlist(item_type=str, min_items=None, max_items=500)]
    organization: Optional[constr(max_length=200)]

    @validator("scheme", pre=True)
    @classmethod
    def check_scheme(cls, value):
        allowed_schemes = ["http", "https"]
        if value not in allowed_schemes:
            raise ValueError(f"The scheme '{value}' must be one of the following {', '.join(allowed_schemes)}")
        return value

    @validator("organization", pre=True)
    @classmethod
    def check_organization(cls, value, values):
        if not value and values.get("provider_type", None) == AZURE_DEVOPS:
            raise ValueError("The organization field needs to be specified for Azure devops vcs instances")
        return value

    @validator("scope", pre=True)
    @classmethod
    def check_scope_and_exceptions(cls, value, values):
        if value and values.get("exceptions", None):
            raise ValueError(
                "You cannot specify bot the scope and exceptions to the scan, only one setting" " is supported."
            )
        return value


class VCSInstanceCreate(VCSInstanceBase):
    pass


class VCSInstanceRead(VCSInstanceBase):
    id_: conint(gt=0)

    @classmethod
    def create_from_db_vcs_instance(cls, db_vcs_instance):
        exceptions = []
        scope = []
        if db_vcs_instance.exceptions:
            exceptions = db_vcs_instance.exceptions.split(",")
        if db_vcs_instance.scope:
            scope = db_vcs_instance.scope.split(",")

        vcs_instance_read = cls(
            id_=db_vcs_instance.id_,
            name=db_vcs_instance.name,
            provider_type=db_vcs_instance.provider_type,
            hostname=db_vcs_instance.hostname,
            port=db_vcs_instance.port,
            scheme=db_vcs_instance.scheme,
            exceptions=exceptions,
            scope=scope,
            organization=db_vcs_instance.organization,
        )

        return vcs_instance_read

    class Config:
        orm_mode = True



# Standard Library
from typing import Generic, Optional, TypeVar

# Third Party
from pydantic import BaseModel, conint
from pydantic.generics import GenericModel

Model = TypeVar("Model", bound=BaseModel)


class FindingCountModel(GenericModel, Generic[Model]):
    """
        Generic encapsulation class for findings count end points to standardize output of the API
        example creation, FindingCountModel[FindingRead](data=db_findings, true_positive=true_positive,
        false_positive=false_positive, not_analyzed=not_analyzed, under_review=under_review,
        clarification_required=clarification_required, total_findings_count=total_findings_count)
    :param Generic[Model]:
        Type of the object in the data list
    """

    data: Optional[Model]
    true_positive: conint(gt=-1)
    false_positive: conint(gt=-1)
    not_analyzed: conint(gt=-1)
    under_review: conint(gt=-1)
    clarification_required: conint(gt=-1)
    total_findings_count: conint(gt=-1)

    class Config:
        orm_mode = True



# Standard Library
import logging
from datetime import datetime, timedelta
from typing import List

# Third Party
from sqlalchemy import extract, func, union
from sqlalchemy.engine import Row
from sqlalchemy.orm import Session
from sqlalchemy.orm.query import Query
from sqlalchemy.sql.expression import literal_column

# First Party
from resc_backend.constants import (
    DEFAULT_RECORDS_PER_PAGE_LIMIT,
    MAX_RECORDS_PER_PAGE_LIMIT,
)
from resc_backend.db.model import (
    DBaudit,
    DBfinding,
    DBrepository,
    DBrule,
    DBruleTag,
    DBscan,
    DBscanFinding,
    DBtag,
    DBVcsInstance,
)
from resc_backend.resc_web_service.crud import scan_finding as scan_finding_crud
from resc_backend.resc_web_service.filters import FindingsFilter
from resc_backend.resc_web_service.schema import finding as finding_schema
from resc_backend.resc_web_service.schema.date_filter import DateFilter
from resc_backend.resc_web_service.schema.finding_status import FindingStatus
from resc_backend.resc_web_service.schema.scan_type import ScanType
from resc_backend.resc_web_service.schema.vcs_provider import VCSProviders

logger = logging.getLogger(__name__)


def patch_finding(db_connection: Session, finding_id: int, finding_update: finding_schema.FindingPatch):
    db_finding = db_connection.query(DBfinding).filter_by(id_=finding_id).first()

    finding_update_dict = finding_update.dict(exclude_unset=True)
    for key in finding_update_dict:
        setattr(db_finding, key, finding_update_dict[key])

    db_connection.commit()
    db_connection.refresh(db_finding)
    return db_finding


def create_findings(db_connection: Session, findings: List[finding_schema.FindingCreate]) -> List[DBfinding]:
    if len(findings) < 1:
        # Function is called with an empty list of findings
        return []

    repository_id = findings[0].repository_id

    # get a list of known / registered findings for this repository
    query = db_connection.query(DBfinding)
    query = query.where(DBfinding.repository_id == repository_id)
    db_repository_findings = query.all()

    # Compare new findings list with findings in the db
    new_findings = findings[:]
    db_findings = []
    for finding in findings:
        for repository_finding in db_repository_findings:
            # Compare based on the unique key in the findings table
            if (
                repository_finding.commit_id == finding.commit_id
                and repository_finding.rule_name == finding.rule_name
                and repository_finding.file_path == finding.file_path
                and repository_finding.line_number == finding.line_number
                and repository_finding.column_start == finding.column_start
                and repository_finding.column_end == finding.column_end
            ):
                # Store the already known finding
                db_findings.append(repository_finding)
                # Remove from the db_repository_findings to increase performance for the next loop
                db_repository_findings.remove(repository_finding)
                # Remove from the to be created findings
                new_findings.remove(finding)
                break
    logger.info(
        f"create_findings repository {repository_id}, Requested: {len(findings)}. "
        f"New findings: {len(new_findings)}. Already in db: {len(db_findings)}"
    )

    db_create_findings = []
    # Map the to be created findings to the DBfinding type object
    for new_finding in new_findings:
        db_create_finding = DBfinding.create_from_finding(new_finding)
        db_create_findings.append(db_create_finding)
    # Store all the to be created findings in the database
    if len(db_create_findings) >= 1:
        db_connection.add_all(db_create_findings)
        db_connection.flush()
        db_connection.commit()
        db_findings.extend(db_create_findings)
    # Return the known findings that are part of the request and the newly created findings
    return db_findings


def get_finding(db_connection: Session, finding_id: int):
    finding = db_connection.query(DBfinding)
    finding = finding.where(DBfinding.id_ == finding_id).first()
    return finding


def get_findings(db_connection: Session, skip: int = 0, limit: int = DEFAULT_RECORDS_PER_PAGE_LIMIT):
    limit_val = MAX_RECORDS_PER_PAGE_LIMIT if limit > MAX_RECORDS_PER_PAGE_LIMIT else limit
    findings = db_connection.query(DBfinding)
    findings = findings.order_by(DBfinding.id_).offset(skip).limit(limit_val).all()
    return findings


def get_scans_findings(
    db_connection,
    scan_ids: List[int],
    skip: int = 0,
    limit: int = DEFAULT_RECORDS_PER_PAGE_LIMIT,
    rules_filter: List[str] = None,
    statuses_filter: List[FindingStatus] = None,
) -> List[DBfinding]:
    """
        Retrieve all finding child objects of a scan object from the database
    :param db_connection:
        Session of the database connection
    :param scan_ids:
        ids of the parent scan object of which to retrieve finding objects
    :param skip:
        integer amount of records to skip to support pagination
    :param limit:
        integer amount of records to return, to support pagination
    :param rules_filter:
        optional, filter on rule name. Is used as a string contains filter
    :param statuses_filter:
        optional, filter on status of findings
    :return: [DBfinding]
        The output will contain a list of DBfinding type objects,
        or an empty list if no finding was found for the given scan_ids
    """
    if len(scan_ids) == 0:
        return []

    limit_val = MAX_RECORDS_PER_PAGE_LIMIT if limit > MAX_RECORDS_PER_PAGE_LIMIT else limit

    query: Query = db_connection.query(DBfinding)
    query = query.join(DBscanFinding, DBscanFinding.finding_id == DBfinding.id_)

    if statuses_filter:
        query = query.join(
            DBaudit,
            (DBaudit.finding_id == DBfinding.id_) & (DBaudit.is_latest == True),  # noqa: E712
            isouter=True,
        )

        if FindingStatus.NOT_ANALYZED.value in statuses_filter:
            query = query.where(
                DBaudit.status.in_(statuses_filter) | (DBaudit.status == None)  # noqa: E711
            )
        else:
            query = query.where(DBaudit.status.in_(statuses_filter))

    query = query.where(DBscanFinding.scan_id.in_(scan_ids))

    if rules_filter:
        query = query.where(DBfinding.rule_name.in_(rules_filter))

    query = query.order_by(DBfinding.id_)
    query = query.offset(skip).limit(limit_val)
    findings = query.all()
    return findings


def get_total_findings_count(db_connection: Session, findings_filter: FindingsFilter = None) -> int:
    """
        Retrieve count of finding records of a given scan
    :param findings_filter:
    :param db_connection:
        Session of the database connection
    :return: total_count
        count of findings
    """

    query = db_connection.query(func.count(DBfinding.id_))

    if findings_filter:
        if findings_filter.finding_statuses:
            query = query.join(
                DBaudit,
                (DBaudit.finding_id == DBfinding.id_) & (DBaudit.is_latest == True),  # noqa: E712
                isouter=True,
            )

        if (
            (findings_filter.vcs_providers and findings_filter.vcs_providers is not None)
            or findings_filter.project_name
            or findings_filter.repository_name
            or findings_filter.start_date_time
            or findings_filter.end_date_time
        ):
            query = query.join(DBscanFinding, DBscanFinding.finding_id == DBfinding.id_)
            query = query.join(DBscan, DBscan.id_ == DBscanFinding.scan_id)
            query = query.join(DBrepository, DBrepository.id_ == DBscan.repository_id)
            query = query.join(DBVcsInstance, DBVcsInstance.id_ == DBrepository.vcs_instance)

        if findings_filter.start_date_time:
            query = query.where(DBscan.timestamp >= findings_filter.start_date_time)
        if findings_filter.end_date_time:
            query = query.where(DBscan.timestamp <= findings_filter.end_date_time)

        if findings_filter.repository_name:
            query = query.where(DBrepository.repository_name == findings_filter.repository_name)

        if findings_filter.vcs_providers and findings_filter.vcs_providers is not None:
            query = query.where(DBVcsInstance.provider_type.in_(findings_filter.vcs_providers))
        if findings_filter.project_name:
            query = query.where(DBrepository.project_key == findings_filter.project_name)
        if findings_filter.rule_names:
            query = query.where(DBfinding.rule_name.in_(findings_filter.rule_names))
        if findings_filter.finding_statuses:
            if FindingStatus.NOT_ANALYZED.value in findings_filter.finding_statuses:
                query = query.where(
                    DBaudit.status.in_(findings_filter.finding_statuses) | (DBaudit.status == None)  # noqa: E711
                )
            else:
                query = query.where(DBaudit.status.in_(findings_filter.finding_statuses))
        if findings_filter.scan_ids and len(findings_filter.scan_ids) == 1:
            query = query.join(DBscanFinding, DBscanFinding.finding_id == DBfinding.id_)
            query = query.where(DBscanFinding.scan_id == findings_filter.scan_ids[0])

        if findings_filter.scan_ids and len(findings_filter.scan_ids) >= 2:
            query = query.join(DBscanFinding, DBscanFinding.finding_id == DBfinding.id_)
            query = query.where(DBscanFinding.scan_id.in_(findings_filter.scan_ids))

    total_count = query.scalar()
    return total_count


def get_findings_by_rule(
    db_connection: Session,
    skip: int = 0,
    limit: int = DEFAULT_RECORDS_PER_PAGE_LIMIT,
    rule_name: str = "",
):
    limit_val = MAX_RECORDS_PER_PAGE_LIMIT if limit > MAX_RECORDS_PER_PAGE_LIMIT else limit
    findings = db_connection.query(DBfinding)
    findings = findings.where(DBfinding.rule_name == rule_name)
    findings = findings.order_by(DBfinding.id_).offset(skip).limit(limit_val).all()
    return findings


def get_distinct_rules_from_findings(
    db_connection: Session,
    scan_id: int = -1,
    finding_statuses: List[FindingStatus] = None,
    vcs_providers: List[VCSProviders] = None,
    project_name: str = "",
    repository_name: str = "",
    start_date_time: datetime = None,
    end_date_time: datetime = None,
    rule_pack_versions: List[str] = None,
) -> List[DBrule]:
    """
        Retrieve distinct rules detected
    :param db_connection:
        Session of the database connection
    :param scan_id:
        Optional filter by the id of a scan
    :param finding_statuses:
        Optional, filter of supported finding statuses
    :param vcs_providers:
        Optional, filter of supported vcs provider types
    :param project_name:
        Optional, filter on project name. Is used as a full string match filter
    :param repository_name:
        optional, filter on repository name. Is used as a string contains filter
    :param start_date_time:
        optional, filter on start date
    :param end_date_time:
        optional, filter on end date
    :param rule_pack_versions:
        optional, filter on rule pack version
    :return: rules
        List of unique rules
    """
    query = db_connection.query(DBfinding.rule_name)

    if (
        vcs_providers or project_name or repository_name or start_date_time or end_date_time or rule_pack_versions
    ) and scan_id < 0:
        query = query.join(DBscanFinding, DBscanFinding.finding_id == DBfinding.id_)
        query = query.join(DBscan, DBscan.id_ == DBscanFinding.scan_id)
        query = query.join(DBrepository, DBrepository.id_ == DBscan.repository_id)
        query = query.join(DBVcsInstance, DBVcsInstance.id_ == DBrepository.vcs_instance)

    if finding_statuses:
        query = query.join(
            DBaudit,
            (DBaudit.finding_id == DBfinding.id_) & (DBaudit.is_latest == True),  # noqa: E712
            isouter=True,
        )

    if scan_id > 0:
        query = query.join(DBscanFinding, DBscanFinding.finding_id == DBfinding.id_)
        query = query.where(DBscanFinding.scan_id == scan_id)
    else:
        if finding_statuses:
            if FindingStatus.NOT_ANALYZED.value in finding_statuses:
                query = query.where(
                    DBaudit.status.in_(finding_statuses) | (DBaudit.status == None)  # noqa: E711
                )
            else:
                query = query.where(DBaudit.status.in_(finding_statuses))

        if vcs_providers:
            query = query.where(DBVcsInstance.provider_type.in_(vcs_providers))

        if project_name:
            query = query.where(DBrepository.project_key == project_name)

        if repository_name:
            query = query.where(DBrepository.repository_name == repository_name)

        if start_date_time:
            query = query.where(DBscan.timestamp >= start_date_time)

        if end_date_time:
            query = query.where(DBscan.timestamp <= end_date_time)

        if rule_pack_versions:
            query = query.where(DBscan.rule_pack.in_(rule_pack_versions))

    rules = query.distinct().order_by(DBfinding.rule_name).all()
    return rules


def get_findings_count_by_status(
    db_connection: Session,
    scan_ids: List[int] = None,
    finding_statuses: List[FindingStatus] = None,
    rule_name: str = "",
):
    """
        Retrieve count of findings based on finding status
    :param db_connection:
        Session of the database connection
    :param scan_ids:
        List of scan ids for which findings should be filtered
    :param finding_statuses:
        finding statuses to filter, type FindingStatus
    :param rule_name:
        rule_name to filter on
    :return: findings_count
        count of findings
    """
    query = db_connection.query(func.count(DBfinding.id_).label("status_count"), DBaudit.status)
    query = query.join(
        DBaudit,
        (DBaudit.finding_id == DBfinding.id_) & (DBaudit.is_latest == True),  # noqa: E712
        isouter=True,
    )

    if scan_ids and len(scan_ids) > 0:
        query = query.join(DBscanFinding, DBscanFinding.finding_id == DBfinding.id_)
        query = query.join(DBscan, DBscan.id_ == DBscanFinding.scan_id)
        query = query.where(DBscan.id_.in_(scan_ids))

    if finding_statuses:
        if FindingStatus.NOT_ANALYZED.value in finding_statuses:
            query = query.where(
                DBaudit.status.in_(finding_statuses) | (DBaudit.status == None)  # noqa: E711
            )
        else:
            query = query.where(DBaudit.status.in_(finding_statuses))
    if rule_name:
        query = query.where(DBfinding.rule_name == rule_name)

    findings_count_by_status = query.group_by(DBaudit.status).all()

    return findings_count_by_status


def get_rule_findings_count_by_status(
    db_connection: Session,
    rule_pack_versions: List[str] = None,
    rule_tags: List[str] = None,
):
    """
        Retrieve count of findings based on rulename and status
    :param db_connection:
        Session of the database connection
    :param rule_pack_versions:
        optional, filter on rule pack version
    :param rule_tags:
        optional, filter on rule tag
    :return: findings_count
        per rulename and status the count of findings
    """
    query = db_connection.query(DBfinding.rule_name, DBaudit.status, func.count(DBfinding.id_))

    max_base_scan_subquery = db_connection.query(
        DBscan.repository_id, func.max(DBscan.id_).label("latest_base_scan_id")
    )
    max_base_scan_subquery = max_base_scan_subquery.where(DBscan.scan_type == ScanType.BASE)
    if rule_pack_versions:
        max_base_scan_subquery = max_base_scan_subquery.where(DBscan.rule_pack.in_(rule_pack_versions))
    max_base_scan_subquery: Query = max_base_scan_subquery.group_by(DBscan.repository_id).subquery()

    query = query.join(DBscanFinding, DBfinding.id_ == DBscanFinding.finding_id)
    query = query.join(
        max_base_scan_subquery,
        DBfinding.repository_id == max_base_scan_subquery.c.repository_id,
    )
    query = query.join(DBscan, (DBscanFinding.scan_id == DBscan.id_))
    query = query.where(DBscan.id_ >= max_base_scan_subquery.c.latest_base_scan_id)

    if rule_tags:
        rule_tag_subquery: Query = db_connection.query(DBruleTag.rule_id).join(DBtag, DBruleTag.tag_id == DBtag.id_)
        if rule_pack_versions:
            rule_tag_subquery = rule_tag_subquery.join(DBrule, DBrule.id_ == DBruleTag.rule_id)
            rule_tag_subquery = rule_tag_subquery.where(DBrule.rule_pack.in_(rule_pack_versions))

        rule_tag_subquery = rule_tag_subquery.where(DBtag.name.in_(rule_tags))
        rule_tag_subquery = rule_tag_subquery.group_by(DBruleTag.rule_id).subquery()

        query = query.join(
            DBrule,
            (DBrule.rule_name == DBfinding.rule_name) & (DBrule.rule_pack == DBscan.rule_pack),
        )
        query = query.join(rule_tag_subquery, DBrule.id_ == rule_tag_subquery.c.rule_id)

    if rule_pack_versions:
        query = query.where(DBscan.rule_pack.in_(rule_pack_versions))

    query = query.join(
        DBaudit,
        (DBaudit.finding_id == DBfinding.id_) & (DBaudit.is_latest == True),  # noqa: E712
        isouter=True,
    )
    query = query.group_by(DBfinding.rule_name, DBaudit.status)
    query = query.order_by(DBfinding.rule_name, DBaudit.status)
    status_counts = query.all()

    rule_count_dict = {}
    for status_count in status_counts:
        rule_count_dict[status_count[0]] = {
            "true_positive": 0,
            "false_positive": 0,
            "not_analyzed": 0,
            "under_review": 0,
            "clarification_required": 0,
            "total_findings_count": 0,
        }

    for status_count in status_counts:
        rule_count_dict[status_count[0]]["total_findings_count"] += status_count[2]
        if status_count[1] == FindingStatus.NOT_ANALYZED.value or status_count[1] is None:
            rule_count_dict[status_count[0]]["not_analyzed"] += status_count[2]
        elif status_count[1] == FindingStatus.FALSE_POSITIVE.value:
            rule_count_dict[status_count[0]]["false_positive"] += status_count[2]
        elif status_count[1] == FindingStatus.TRUE_POSITIVE.value:
            rule_count_dict[status_count[0]]["true_positive"] += status_count[2]
        elif status_count[1] == FindingStatus.UNDER_REVIEW.value:
            rule_count_dict[status_count[0]]["under_review"] += status_count[2]
        elif status_count[1] == FindingStatus.CLARIFICATION_REQUIRED.value:
            rule_count_dict[status_count[0]]["clarification_required"] += status_count[2]

    return rule_count_dict


def get_findings_count_by_time(
    db_connection: Session,
    date_type: DateFilter,
    start_date_time: datetime = None,
    end_date_time: datetime = None,
    skip: int = 0,
    limit: int = DEFAULT_RECORDS_PER_PAGE_LIMIT,
):
    """
        Retrieve count based on date_type
    :param db_connection:
        Session of the database connection
    :param date_type:
        required, filter on time_type
    :param start_date_time:
        optional, filter on start date
    :param end_date_time:
        optional, filter on end date
    :param skip:
        integer amount of records to skip to support pagination
    :param limit:
        integer amount of records to return, to support pagination
    """
    if date_type == DateFilter.MONTH:
        query = db_connection.query(
            extract("year", DBscan.timestamp),
            extract("month", DBscan.timestamp),
            func.count(DBscanFinding.finding_id),
        )
    elif date_type == DateFilter.WEEK:
        query = db_connection.query(
            extract("year", DBscan.timestamp),
            extract("week", DBscan.timestamp),
            func.count(DBscanFinding.finding_id),
        )
    elif date_type == DateFilter.DAY:
        query = db_connection.query(
            extract("year", DBscan.timestamp),
            extract("month", DBscan.timestamp),
            extract("day", DBscan.timestamp),
            func.count(DBscanFinding.finding_id),
        )

    query: Query = query.join(DBscanFinding, DBscanFinding.scan_id == DBscan.id_)

    if start_date_time:
        query = query.where(DBscan.timestamp >= start_date_time)
    if end_date_time:
        query = query.where(DBscan.timestamp <= end_date_time)

    if date_type == DateFilter.MONTH:
        query = query.group_by(extract("year", DBscan.timestamp), extract("month", DBscan.timestamp))
        query = query.order_by(extract("year", DBscan.timestamp), extract("month", DBscan.timestamp))
    elif date_type == DateFilter.WEEK:
        query = query.group_by(extract("year", DBscan.timestamp), extract("week", DBscan.timestamp))
        query = query.order_by(extract("year", DBscan.timestamp), extract("week", DBscan.timestamp))
    elif date_type == DateFilter.DAY:
        query = query.group_by(
            extract("year", DBscan.timestamp),
            extract("month", DBscan.timestamp),
            extract("day", DBscan.timestamp),
        )
        query = query.order_by(
            extract("year", DBscan.timestamp),
            extract("month", DBscan.timestamp),
            extract("day", DBscan.timestamp),
        )

    finding_count = query.offset(skip).limit(limit).all()
    return finding_count


def get_findings_count_by_time_total(
    db_connection: Session,
    date_type: DateFilter,
    start_date_time: datetime = None,
    end_date_time: datetime = None,
):
    """
        Retrieve total count on date_type
    :param db_connection:
        Session of the database connection
    :param date_type:
        required, filter on time_type
    :param start_date_time:
        optional, filter on start date
    :param end_date_time:
        optional, filter on end date
    """
    if date_type == DateFilter.MONTH:
        query: Query = db_connection.query(extract("year", DBscan.timestamp), extract("month", DBscan.timestamp))
    elif date_type == DateFilter.WEEK:
        query: Query = db_connection.query(extract("year", DBscan.timestamp), extract("week", DBscan.timestamp))
    elif date_type == DateFilter.DAY:
        query: Query = db_connection.query(
            extract("year", DBscan.timestamp),
            extract("month", DBscan.timestamp),
            extract("day", DBscan.timestamp),
        )

    query = query.join(DBscanFinding, DBscan.id_ == DBscanFinding.scan_id)

    if start_date_time:
        query = query.where(DBscan.timestamp >= start_date_time)
    if end_date_time:
        query = query.where(DBscan.timestamp <= end_date_time)

    query = query.distinct()

    result = query.count()
    return result


def get_distinct_rules_from_scans(db_connection: Session, scan_ids: List[int] = None) -> List[DBrule]:
    """
        Retrieve distinct rules detected
    :param db_connection:
        Session of the database connection
    :param scan_ids:
        List of scan ids
    :return: rules
        List of unique rules
    """
    query = db_connection.query(DBfinding.rule_name)

    if scan_ids:
        query = query.join(DBscanFinding, DBscanFinding.finding_id == DBfinding.id_)
        query = query.where(DBscanFinding.scan_id.in_(scan_ids))

    query = query.distinct()
    query = query.order_by(DBfinding.rule_name)
    rules = query.all()
    return rules


def delete_finding(db_connection: Session, finding_id: int, delete_related: bool = False):
    """
        Delete a finding object
    :param db_connection:
        Session of the database connection
    :param finding_id:
        id of the finding to be deleted
    :param delete_related:
        if related records need to be deleted
    """
    if delete_related:
        scan_finding_crud.delete_scan_finding(db_connection, finding_id=finding_id)

    db_connection.query(DBfinding).where(DBfinding.id_ == finding_id).delete(synchronize_session=False)
    db_connection.commit()


def delete_findings_by_repository_id(db_connection: Session, repository_id: int):
    """
        Delete findings for a given repository
    :param db_connection:
        Session of the database connection
    :param repository_id:
        id of the repository
    """
    db_connection.query(DBfinding).where(DBfinding.repository_id == repository_id).delete(synchronize_session=False)
    db_connection.commit()


def delete_findings_by_vcs_instance_id(db_connection: Session, vcs_instance_id: int):
    """
        Delete findings for a given vcs instance
    :param db_connection:
        Session of the database connection
    :param vcs_instance_id:
        id of the vcs instance
    """
    db_connection.query(DBfinding).where(
        DBfinding.repository_id == DBrepository.id_,
        DBrepository.vcs_instance == DBVcsInstance.id_,
        DBVcsInstance.id_ == vcs_instance_id,
    ).delete(synchronize_session=False)
    db_connection.commit()


def _get_iso_date_now_diff_week(week: int) -> datetime:
    """
        Taking now() computes the timestamp associated to the
        ISO calendar year.

    Args:
        week (int): Number of weeks to substract to the timestamp.

    Returns:
        datetime: shifted iso time stamp for easier computations.
    """
    current_utc_time = datetime.utcnow()
    current_iso = current_utc_time.isocalendar()
    current_iso_year = current_iso[0]
    current_iso_week = current_iso[1]
    # We use 6 (Saturday) to include the current week when 0
    # rather than doing a shift from current day.
    last_nth_week_date_time = datetime.fromisocalendar(current_iso_year, current_iso_week, 6) - timedelta(weeks=week)
    return last_nth_week_date_time


def _max_base_scan_subequery(db_connection: Session, last_nth_week_date_time: datetime) -> Query:
    """
        Creates a subquery given a cut-off date to have the max base scan
        per repository

    Args:
        db_connection (Session): Session to generate the subquery
        last_nth_week_date_time (datetime): max cut-off date

    Returns:
        Query: subquery to have the max base scan under a cut-off date
    """
    subquery: Query = db_connection.query(func.max(DBscan.id_).label("scan_id"), DBscan.repository_id)
    subquery = subquery.where(DBscan.timestamp <= last_nth_week_date_time)
    subquery = subquery.where(DBscan.scan_type == ScanType.BASE)
    subquery = subquery.group_by(DBscan.repository_id)
    return subquery.subquery()


def _max_audit_subequery(db_connection: Session, last_nth_week_date_time: datetime) -> Query:
    """
        Creates a subquery given a cut-off date to have the last audit
        per finding

    Args:
        db_connection (Session): Session to generate the subquery
        last_nth_week_date_time (datetime): max cut-off date

    Returns:
        Query: subquery to have the last audit under a cut-off date
    """
    subquery: Query = db_connection.query(func.max(DBaudit.id_).label("audit_id"), DBaudit.finding_id)
    subquery = subquery.where(DBaudit.timestamp < last_nth_week_date_time)
    subquery = subquery.group_by(DBaudit.finding_id)
    return subquery.subquery()


def get_finding_audit_status_count_over_time(db_connection: Session, status: FindingStatus, weeks: int = 13) -> dict:
    """
        Retrieve count of true positive findings over time for given weeks
    :param db_connection:
        Session of the database connection
    :param status:
        mandatory, status for which to get the audit counts over time
    :param weeks:
        optional, filter on last n weeks, default 13
    :return: true_positive_count_over_time
        list of rows containing finding statuses count over time per week
    """
    all_tables = []
    for week in range(0, weeks):
        last_nth_week_date_time = _get_iso_date_now_diff_week(week)
        max_audit_subquery = _max_audit_subequery(db_connection, last_nth_week_date_time)

        query = db_connection.query(
            literal_column(str(last_nth_week_date_time.isocalendar()[0])).label("year"),
            literal_column(str(last_nth_week_date_time.isocalendar()[1])).label("week"),
            DBVcsInstance.provider_type.label("provider_type"),
            func.count(DBaudit.id_).label("finding_count"),
        )
        query = query.join(max_audit_subquery, max_audit_subquery.c.audit_id == DBaudit.id_)
        query = query.join(DBfinding, DBfinding.id_ == DBaudit.finding_id)
        query = query.join(DBrepository, DBrepository.id_ == DBfinding.repository_id)
        query = query.join(DBVcsInstance, DBVcsInstance.id_ == DBrepository.vcs_instance)
        query = query.where(DBaudit.status == status)
        query = query.group_by(DBVcsInstance.provider_type)

        all_tables.append(query)

    # union
    unioned_query = union(*all_tables)
    status_count_over_time = db_connection.execute(unioned_query).all()
    return status_count_over_time


def get_finding_count_by_vcs_provider_over_time(db_connection: Session, weeks: int = 13) -> list[Row]:
    """
        Retrieve count findings over time for given weeks
    :param db_connection:
        Session of the database connection
    :param weeks:
        optional, filter on last n weeks, default 13
    :return: count_over_time
        list of rows containing finding count over time per week
    """
    all_tables = []
    for week in range(0, weeks):
        last_nth_week_date_time = _get_iso_date_now_diff_week(week)

        max_base_scan = _max_base_scan_subequery(db_connection, last_nth_week_date_time)

        query = db_connection.query(
            literal_column(str(last_nth_week_date_time.isocalendar()[0])).label("year"),
            literal_column(str(last_nth_week_date_time.isocalendar()[1])).label("week"),
            DBVcsInstance.provider_type.label("provider_type"),
            func.count(DBfinding.id_).label("finding_count"),
        )
        query = query.join(DBscanFinding, DBfinding.id_ == DBscanFinding.finding_id)
        query = query.join(DBscan, DBscan.id_ == DBscanFinding.scan_id)
        query = query.join(max_base_scan, max_base_scan.c.repository_id == DBscan.repository_id)
        query = query.where(DBscan.id_ >= max_base_scan.c.scan_id)
        query = query.where(DBscan.timestamp <= last_nth_week_date_time)
        query = query.join(DBrepository, DBrepository.id_ == DBscan.repository_id)
        query = query.join(DBVcsInstance, DBVcsInstance.id_ == DBrepository.vcs_instance)
        query = query.group_by(DBVcsInstance.provider_type)

        all_tables.append(query)

    # union
    unioned_query = union(*all_tables)
    count_over_time = db_connection.execute(unioned_query).all()
    return count_over_time


def get_un_triaged_finding_count_by_vcs_provider_over_time(db_connection: Session, weeks: int = 13) -> list[Row]:
    """
        Retrieve count of un triaged findings over time for given weeks
    :param db_connection:
        Session of the database connection
    :param weeks:
        optional, filter on last n weeks, default 13
    :return: count_over_time
        list of rows containing un triaged findings count over time per week
    """
    all_tables = []
    for week in range(0, weeks):
        last_nth_week_date_time = _get_iso_date_now_diff_week(week)
        max_base_scan = _max_base_scan_subequery(db_connection, last_nth_week_date_time)
        max_audit_subquery = _max_audit_subequery(db_connection, last_nth_week_date_time)

        query = db_connection.query(
            literal_column(str(last_nth_week_date_time.isocalendar()[0])).label("year"),
            literal_column(str(last_nth_week_date_time.isocalendar()[1])).label("week"),
            DBVcsInstance.provider_type.label("provider_type"),
            func.count(DBfinding.id_).label("finding_count"),
        )
        query = query.join(DBscanFinding, DBfinding.id_ == DBscanFinding.finding_id)
        query = query.join(DBscan, DBscan.id_ == DBscanFinding.scan_id)
        query = query.join(DBrepository, DBrepository.id_ == DBscan.repository_id)
        query = query.join(DBVcsInstance, DBVcsInstance.id_ == DBrepository.vcs_instance)
        query = query.join(max_base_scan, max_base_scan.c.repository_id == DBscan.repository_id)
        query = query.where(DBscan.id_ >= max_base_scan.c.scan_id)
        query = query.where(DBscan.timestamp <= last_nth_week_date_time)
        query = query.join(
            max_audit_subquery,
            max_audit_subquery.c.finding_id == DBfinding.id_,
            isouter=True,
        )
        query = query.join(
            DBaudit,
            (DBaudit.finding_id == DBfinding.id_) & (DBaudit.id_ == max_audit_subquery.c.audit_id),
            isouter=True,
        )
        query = query.where(
            (DBaudit.id_ == None) | (DBaudit.status == FindingStatus.NOT_ANALYZED.value)  # noqa: E711
        )
        query = query.group_by(DBVcsInstance.provider_type)
        all_tables.append(query)

    # union
    unioned_query = union(*all_tables)
    count_over_time = db_connection.execute(unioned_query).all()
    return count_over_time



# Standard Library
from typing import Generic, TypeVar

# Third Party
# pylint: disable=no-name-in-module
from pydantic import BaseModel, conint, conlist
from pydantic.generics import GenericModel

Model = TypeVar("Model", bound=BaseModel)


class PaginationModel(GenericModel, Generic[Model]):
    """
        Generic encapsulation class for paginated endpoints to standardize output of the API
        example creation, PaginationModel[FindingRead](data=db_findings, total=total, limit=limit, skip=skip)
    :param Generic[Model]:
        Type of the object in the data list
    """

    # data: List[Model]
    data: conlist(item_type=Model, min_items=None, max_items=500)
    total: conint(gt=-1)
    limit: conint(gt=-1)
    skip: conint(gt=-1)

    class Config:
        orm_mode = True



# Standard Library
import logging.config
import sysconfig
from os import path

# Third Party
import pkg_resources

logger = logging.getLogger(__name__)


def get_logging_settings_path():
    if path.isfile(sysconfig.get_path("purelib") + "/resc"):
        base_dir = sysconfig.get_path("purelib") + "/resc"
    else:
        base_dir = path.dirname(__file__)

    return base_dir + "/static/logging.ini"


def initialise_logs(log_file_path: str, debug=True):
    logging_ini_file = get_logging_settings_path()
    logging.config.fileConfig(
        logging_ini_file,
        defaults={"log_file_path": log_file_path},
        disable_existing_loggers=False,
    )
    logger_config = logging.getLogger("root")
    if int(debug) == 1:
        logger_config.setLevel(logging.DEBUG)
    else:
        logger_config.setLevel(logging.INFO)
    return logger_config


def get_package_version():
    try:
        version = pkg_resources.get_distribution("resc-backend").version
    except pkg_resources.DistributionNotFound:
        version = "1.0.0"
    return version



# Standard Library
import datetime

# Third Party
from pydantic import BaseModel, conint, constr

# First Party
from resc_backend.resc_web_service.schema.scan_type import ScanType


class ScanBase(BaseModel):
    scan_type: ScanType = ScanType.BASE
    last_scanned_commit: constr(min_length=1, max_length=100)
    timestamp: datetime.datetime
    increment_number: conint(gt=-1) = 0
    rule_pack: constr(max_length=100)


class ScanCreate(ScanBase):
    repository_id: conint(gt=0)

    @classmethod
    def create_from_base_class(cls, base_object: ScanBase, repository_id: int):
        return cls(**(dict(base_object)), repository_id=repository_id)


class Scan(ScanBase):
    pass


class ScanRead(ScanCreate):
    id_: conint(gt=0)

    class Config:
        orm_mode = True



# Standard Library
import logging
from typing import List

# Third Party
from sqlalchemy.orm import Session
from sqlalchemy.orm.query import Query

# First Party
from resc_backend.db.model import DBrule, DBruleAllowList, DBrulePack, DBscan
from resc_backend.resc_web_service.schema import (
    rule_allow_list as rule_allow_list_schema,
)
from resc_backend.resc_web_service.schema.rule import RuleCreate, RuleRead

logger = logging.getLogger(__name__)


def get_rules_by_scan_id(db_connection: Session, scan_id: int) -> List[RuleRead]:
    """
        Get rules by scan id
    :param db_connection:
        Session of the database connection
    :param scan_id:
        scan id for which rules need to be fetched
    :return: List[RuleRead]
        The output contains list of rules
    """
    rule_query = db_connection.query(DBrule)
    rule_query = rule_query.join(DBscan, DBscan.rule_pack == DBrule.rule_pack)
    rule_query = rule_query.where(DBscan.id_ == scan_id)
    rules: List[RuleRead] = rule_query.all()
    return rules


def create_rule_allow_list(db_connection: Session, rule_allow_list: rule_allow_list_schema.RuleAllowList):
    """
        Create rule allow list in database
    :param db_connection:
        Session of the database connection
    :param rule_allow_list:
        RuleAllowList object to be created
    """
    db_rule_allow_list = DBruleAllowList(
        description=rule_allow_list.description,
        regexes=rule_allow_list.regexes,
        paths=rule_allow_list.paths,
        commits=rule_allow_list.commits,
        stop_words=rule_allow_list.stop_words,
    )
    db_connection.add(db_rule_allow_list)
    db_connection.commit()
    db_connection.refresh(db_rule_allow_list)
    return db_rule_allow_list


def create_rule(db_connection: Session, rule: RuleCreate):
    """
        Create rule in database
    :param db_connection:
        Session of the database connection
    :param rule:
        RuleCreate object to be created
    """
    db_rule = DBrule(
        rule_name=rule.rule_name,
        description=rule.description,
        entropy=rule.entropy,
        secret_group=rule.secret_group,
        regex=rule.regex,
        path=rule.path,
        keywords=rule.keywords,
        rule_pack=rule.rule_pack,
        allow_list=rule.allow_list,
    )
    db_connection.add(db_rule)
    db_connection.commit()
    db_connection.refresh(db_rule)
    return db_rule


def get_rules_by_rule_pack_version(db_connection: Session, rule_pack_version: str) -> List[str]:
    """
        Fetch rules by rule pack version
    :param db_connection:
        Session of the database connection
    :param rule_pack_version:
        rule pack version
    :return: List[str]
        The output contains list of strings of global allow list
    """
    query: Query = db_connection.query(
        DBrule.id_,
        DBrule.rule_pack,
        DBrule.rule_name,
        DBrule.entropy,
        DBrule.secret_group,
        DBrule.regex,
        DBrule.path,
        DBrule.keywords,
        DBruleAllowList.description,
        DBruleAllowList.regexes,
        DBruleAllowList.paths,
        DBruleAllowList.commits,
        DBruleAllowList.stop_words,
    )

    query = query.join(DBrulePack, DBrulePack.version == DBrule.rule_pack)
    query = query.join(DBruleAllowList, DBruleAllowList.id_ == DBrule.allow_list, isouter=True)
    query = query.where(DBrule.rule_pack == rule_pack_version)
    query = query.order_by(DBrule.id_)
    db_rules = query.all()

    return db_rules


def get_global_allow_list_by_rule_pack_version(db_connection: Session, rule_pack_version: str) -> List[str]:
    """
        Retrieve global allow list by rule pack version
    :param db_connection:
        Session of the database connection
    :param rule_pack_version:
        rule pack version
    :return: List[str]
        The output contains list of strings of global allow list
    """
    query: Query = db_connection.query(
        DBrulePack.version,
        DBruleAllowList.description,
        DBruleAllowList.regexes,
        DBruleAllowList.paths,
        DBruleAllowList.commits,
        DBruleAllowList.stop_words,
    )
    query = query.join(DBruleAllowList, DBruleAllowList.id_ == DBrulePack.global_allow_list)
    query = query.where(DBrulePack.version == rule_pack_version)
    query = query.order_by(DBruleAllowList.id_)
    db_global_allow_list = query.first()

    return db_global_allow_list



# Third Party
from setuptools import setup


def get_required_packages():
    with open("requirements.txt") as f:
        return f.read().splitlines()


setup(install_requires=get_required_packages())



# Standard Library
import logging.config

# Third Party
from fastapi import Depends, FastAPI
from fastapi.middleware.cors import CORSMiddleware
from starlette.responses import RedirectResponse
from starlette.status import HTTP_302_FOUND
from tenacity import RetryError

# First Party
from resc_backend.common import get_package_version
from resc_backend.constants import RWS_VERSION_PREFIX
from resc_backend.db.connection import Session, engine
from resc_backend.helpers.environment_wrapper import validate_environment
from resc_backend.resc_web_service.configuration import (
    AUTHENTICATION_REQUIRED,
    CORS_ALLOWED_DOMAINS,
    ENABLE_CORS,
    WEB_SERVICE_ENV_VARS,
)
from resc_backend.resc_web_service.dependencies import (
    check_db_initialized,
    requires_auth,
    requires_no_auth,
    add_security_headers,
)
from resc_backend.resc_web_service.endpoints import (
    common,
    detailed_findings,
    findings,
    health,
    metrics,
    repositories,
    rules,
    rule_packs,
    scans,
    vcs_instances,
)
from resc_backend.resc_web_service.helpers.exception_handler import (
    add_exception_handlers,
)
from resc_backend.resc_web_service.cache_manager import CacheManager

# Check and load environment variables
env_variables = validate_environment(WEB_SERVICE_ENV_VARS)


def generate_logger_config(log_file_path, debug=True):
    """A function to generate the global logger config dictionary

    Arguments:
        log_file_path {string} -- Path where the logs are to be stored

    Keyword Arguments:
        debug {bool} -- Whether the logging level should be set to DEBUG or INFO (default: {True})

    Returns:
        Dict -- A dictionary containing the logger configuration
    """

    logging_level = "DEBUG" if debug else "INFO"
    logging_config = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "generic-log-formatter": {"format": "[%(levelname)s] [%(name)s] [%(asctime)s] %(message)s"},
        },
        "handlers": {
            "console": {
                "level": logging_level,
                "class": "logging.StreamHandler",
                "formatter": "generic-log-formatter",
            },
            "file": {
                "level": logging_level,
                "class": "logging.handlers.RotatingFileHandler",
                "formatter": "generic-log-formatter",
                "filename": log_file_path,
                "maxBytes": 100 * 1024 * 1024,
                "backupCount": 5,
            },
        },
        "loggers": {
            "": {
                "handlers": ["console", "file"],
                "level": logging_level,
                "propagate": True,
            },
        },
    }

    return logging_config


logging.config.dictConfig(generate_logger_config("local_logs.log"))
logger = logging.getLogger(__name__)
tags_metadata = [
    {"name": "health", "description": "Checks health for API"},
    {"name": "resc-common", "description": "Manage common information"},
    {"name": "resc-rules", "description": "Manage rule information"},
    {"name": "resc-rule-packs", "description": "Manage rule pack information"},
    {"name": "resc-repositories", "description": "Manage repository information"},
    {"name": "resc-scans", "description": "Manage scan information"},
    {"name": "resc-findings", "description": "Manage findings information"},
    {"name": "resc-vcs-instances", "description": "Manage vcs instance information"},
    {"name": "resc-metrics", "description": "Retrieve metrics"},
]

# Check if authentication is required for api endpoints
auth_disabled = env_variables[AUTHENTICATION_REQUIRED].lower() in ["false"]
AUTH = [Depends(requires_no_auth)] if auth_disabled else [Depends(requires_auth)]

app = FastAPI(
    title="Repository Scanner (RESC)",
    description="RESC API helps you to perform several operations upon findings "
    "obtained from multiple source code repositories.",
    version=get_package_version(),
    openapi_tags=tags_metadata,
)

if env_variables[ENABLE_CORS].lower() in ["true"]:
    origins = env_variables[CORS_ALLOWED_DOMAINS].split(", ")
    app.add_middleware(
        CORSMiddleware,
        allow_origins=origins,
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

app.include_router(health.router, prefix=RWS_VERSION_PREFIX)
app.include_router(common.router, prefix=RWS_VERSION_PREFIX, dependencies=AUTH)
app.include_router(rules.router, prefix=RWS_VERSION_PREFIX, dependencies=AUTH)
app.include_router(rule_packs.router, prefix=RWS_VERSION_PREFIX, dependencies=AUTH)
app.include_router(findings.router, prefix=RWS_VERSION_PREFIX, dependencies=AUTH)
app.include_router(detailed_findings.router, prefix=RWS_VERSION_PREFIX, dependencies=AUTH)
app.include_router(repositories.router, prefix=RWS_VERSION_PREFIX, dependencies=AUTH)
app.include_router(scans.router, prefix=RWS_VERSION_PREFIX, dependencies=AUTH)
app.include_router(vcs_instances.router, prefix=RWS_VERSION_PREFIX, dependencies=AUTH)
app.include_router(metrics.router, prefix=RWS_VERSION_PREFIX, dependencies=AUTH)

# Apply the security headers to the app in the form of middleware
app.middleware("http")(add_security_headers)

# Add exception handlers
add_exception_handlers(app=app)


@app.on_event("startup")
def app_startup():
    CacheManager.initialize_cache(env_variables=env_variables)
    try:
        _ = Session(bind=engine)
        check_db_initialized()

        logger.info("Database is connected, expected table(s) found")
    except RetryError as exc:
        raise SystemExit("Error while connecting to the database, retry timed out") from exc


@app.on_event("shutdown")
async def app_shutdown():
    await CacheManager.clear_all_cache()


@app.get("/")
def view_docs():
    return RedirectResponse(url="/docs", status_code=HTTP_302_FOUND)
